import os
import sys
import random
import string
import asyncio
import logging
import argparse

import naz

from .utils import sig, load

os.environ["PYTHONASYNCIODEBUG"] = "1"


def make_parser():
    """
    this is abstracted into its own method so that it is easier to test it.
    """
    parser = argparse.ArgumentParser(
        prog="naz",
        description="""naz is an async SMPP client.
                example usage:
                naz-cli \
                --client dotted.path.to.naz.Client.instance
                """,
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=naz.__version__.about["__version__"]),
        help="The currently installed naz version.",
    )
    parser.add_argument(
        "--client",
        required=True,
        help="The dotted path to a `naz.Client` instance. \
        eg: --client dotted.path.to.a.naz.Client.class.instance",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        required=False,
        default=False,
        help="""Whether we want to do a dry-run of the naz cli.
        This is typically only used by developers who are developing naz.
        eg: --dry-run""",
    )
    return parser


def main():
    """
    entrypoint of naz-cli app.
    """
    parser = make_parser()
    args = parser.parse_args()
    dry_run = args.dry_run
    client = args.client

    _client_id = "".join(random.choices(string.ascii_uppercase + string.digits, k=17))
    logger = naz.log.SimpleLogger("naz.cli")
    try:
        logger.log(logging.INFO, "\n\n\t {} \n\n".format("Naz: the SMPP client."))
        logger.log(
            logging.INFO, {"event": "naz.cli.main", "stage": "start", "client_id": _client_id}
        )
        client = load.load_class(dotted_path=client, logger=logger)
        if dry_run:
            logger.log(
                logging.WARN,
                "\n\n\t {} \n\n".format(
                    "Naz: Caution; You have activated dry-run, naz may not behave correctly."
                ),
            )

        if not isinstance(client, naz.Client):
            e = ValueError(
                """`client` should be of type:: `naz.Client` You entered: {0}""".format(
                    type(client)
                )
            )
            logger.log(logging.ERROR, {"event": "naz.cli.main", "stage": "end", "error": str(e)})
            sys.exit(77)

        if dry_run:
            return
        # call naz api
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            async_main(client=client, logger=logger, loop=loop, dry_run=dry_run)
        )
    except Exception as e:
        logger.log(logging.ERROR, {"event": "naz.cli.main", "stage": "end", "error": str(e)})
        sys.exit(77)
    finally:
        logger.log(logging.INFO, {"event": "naz.cli.main", "stage": "end"})


async def async_main(
    client: naz.Client,
    logger: naz.log.SimpleLogger,
    loop: asyncio.events.AbstractEventLoop,
    dry_run: bool,
):
    # connect & bind to the SMSC host
    await client.connect()
    await client.tranceiver_bind()

    # send any queued messages to SMSC, read any data from SMSC and continually check the state of the SMSC
    tasks = asyncio.gather(
        client.dequeue_messages(TESTING=dry_run),
        client.receive_data(TESTING=dry_run),
        client.enquire_link(TESTING=dry_run),
        sig._signal_handling(logger=logger, client=client, loop=loop),
        loop=loop,
    )
    await tasks


if __name__ == "__main__":
    main()



import os
import struct
import random
import socket
import string
import typing
import asyncio
import logging

# pytype: disable=pyi-error
from . import q
from . import hooks
from . import log
from . import nazcodec
from . import sequence
from . import throttle
from . import correlater
from . import ratelimiter

from .state import (
    SmppSessionState,
    SmppCommand,
    CommandStatus,
    SmppCommandStatus,
    SmppDataCoding,
    SmppOptionalTag,
)

# pytype: disable=pyi-error


class Client:
    """
    The SMPP client that will interact with SMSC/server.

    Example declaration:

    .. highlight:: python
    .. code-block:: python

        import os
        import asyncio
        import naz

        outboundqueue = naz.q.SimpleOutboundQueue(maxsize=1000)
        client = naz.Client(
                smsc_host="127.0.0.1",
                smsc_port=2775,
                system_id="smppclient1",
                password=os.getenv("password", "password"),
                outboundqueue=outboundqueue,
            )

        loop = asyncio.get_event_loop()
        # connect to the SMSC host
        _, _ = loop.run_until_complete(client.connect())
        # bind to SMSC as a tranceiver
        loop.run_until_complete(client.tranceiver_bind())

        # read any data from SMSC, send any queued messages to SMSC and continually check the state of the SMSC
        tasks = asyncio.gather(
            client.dequeue_messages(),
            client.receive_data(),
            client.enquire_link(),
            loop=loop,
        )
        loop.run_until_complete(tasks)

    """

    def __init__(
        self,
        smsc_host: str,
        smsc_port: int,
        system_id: str,
        password: str,
        outboundqueue: q.BaseOutboundQueue,
        client_id: typing.Union[None, str] = None,
        system_type: str = "",
        addr_ton: int = 0,
        addr_npi: int = 0,
        address_range: str = "",
        encoding: str = "gsm0338",
        interface_version: int = 34,
        service_type: str = "CMT",  # section 5.2.11
        source_addr_ton: int = 0x00000001,  # section 5.2.5
        source_addr_npi: int = 0x00000001,
        dest_addr_ton: int = 0x00000001,
        dest_addr_npi: int = 0x00000001,
        # xxxxxx00 store-and-forward
        # xx0010xx Short Message contains ESME Delivery Acknowledgement
        # 00xxxxxx No specific features selected
        esm_class: int = 0b00000011,  # section 5.2.12
        protocol_id: int = 0x00000000,
        priority_flag: int = 0x00000000,
        schedule_delivery_time: str = "",
        validity_period: str = "",
        # xxxxxx01 SMSC Delivery Receipt requested where final delivery outcome is delivery success or failure
        # xxxx01xx SME Delivery Acknowledgement requested
        # xxx0xxxx No Intermediate notification requested
        # all other values reserved
        registered_delivery: int = 0b00000001,  # see section 5.2.17
        replace_if_present_flag: int = 0x00000000,
        sm_default_msg_id: int = 0x00000000,
        enquire_link_interval: float = 55.00,
        logger: typing.Union[None, log.BaseLogger] = None,
        loglevel: str = "INFO",
        log_metadata: typing.Union[None, dict] = None,
        codec_class: typing.Union[None, nazcodec.BaseNazCodec] = None,
        codec_errors_level: str = "strict",
        rateLimiter: typing.Union[None, ratelimiter.BaseRateLimiter] = None,
        hook: typing.Union[None, hooks.BaseHook] = None,
        sequence_generator: typing.Union[None, sequence.BaseSequenceGenerator] = None,
        throttle_handler: typing.Union[None, throttle.BaseThrottleHandler] = None,
        correlation_handler: typing.Union[None, correlater.BaseCorrelater] = None,
        drain_duration: float = 8.00,
        socket_timeout: float = 30.0,
    ) -> None:
        """
        Parameters:
            smsc_host:	the IP address(or domain name) of the SMSC gateway/server
            smsc_port:	the port at which SMSC is listening on
            system_id:	Identifies the ESME system requesting to bind as a transceiver with the SMSC.
            password:	The password to be used by the SMSC to authenticate the ESME requesting to bind.
            system_type:	Identifies the type of ESME system requesting to bind with the SMSC.
            addr_ton:	Type of Number of the ESME address.
            addr_npi:	Numbering Plan Indicator (NPI) for ESME address(es) served via this SMPP transceiver session
            address_range:	A single ESME address or a range of ESME addresses served via this SMPP transceiver session.
            interface_version:	Indicates the version of the SMPP protocol supported by the ESME.
            service_type:	Indicates the SMS Application service associated with the message
            source_addr_ton:	Type of Number of message originator.
            source_addr_npi:	Numbering Plan Identity of message originator.
            dest_addr_ton:	Type of Number for destination.
            dest_addr_npi:	Numbering Plan Identity of destination
            esm_class:	Indicates Message Mode & Message Type.
            protocol_id:	Protocol Identifier. Network specific field.
            priority_flag:	Designates the priority level of the message.
            schedule_delivery_time:	The short message is to be scheduled by the SMSC for delivery.
            validity_period:	The validity period of this message.
            registered_delivery:	Indicator to signify if an SMSC delivery receipt or an SME acknowledgement is required.
            replace_if_present_flag:	Flag indicating if submitted message should replace an existing message.
            sm_default_msg_id:	Indicates the short message to send from a list of predefined (‘canned’) short messages stored on the SMSC
            encoding:	encoding1 used to encode messages been sent to SMSC
            sequence_generator:	python class instance used to generate sequence_numbers
            outboundqueue:	python class instance implementing some queueing mechanism. \
                messages to be sent to SMSC are queued using the said mechanism before been sent
            client_id:	a unique string identifying a naz client class instance
            logger: python class instance to be used for logging
            loglevel:	the level at which to log
            log_metadata: metadata that will be included in all log statements
            codec_class: python class instance to be used to encode/decode messages
            codec_errors_level:	same meaning as the errors argument to pythons' encode method as defined here
            enquire_link_interval:	time in seconds to wait before sending an enquire_link request to SMSC to check on its status
            rateLimiter: python class instance implementing rate limitation
            hook: python class instance implemeting functionality/hooks to be called by naz \
                just before sending request to SMSC and just after getting response from SMSC
            throttle_handler: python class instance implementing functionality of what todo when naz starts getting throttled responses from SMSC
            correlation_handler: A python class instance that naz uses to store relations between \
                SMPP sequence numbers and user applications' log_id's and/or hook_metadata.
            drain_duration: duration in seconds that `naz` will wait for after receiving a termination signal.
            socket_timeout: duration that `naz` will wait, for socket/connection related activities with SMSC, before timing out

        Raises:
            NazClientError: raised if there’s an error instantiating a naz Client.
        """
        self._validate_client_args(
            smsc_host=smsc_host,
            smsc_port=smsc_port,
            system_id=system_id,
            password=password,
            outboundqueue=outboundqueue,
            client_id=client_id,
            system_type=system_type,
            addr_ton=addr_ton,
            addr_npi=addr_npi,
            address_range=address_range,
            encoding=encoding,
            interface_version=interface_version,
            service_type=service_type,
            source_addr_ton=source_addr_ton,
            source_addr_npi=source_addr_npi,
            dest_addr_ton=dest_addr_ton,
            dest_addr_npi=dest_addr_npi,
            esm_class=esm_class,
            protocol_id=protocol_id,
            priority_flag=priority_flag,
            schedule_delivery_time=schedule_delivery_time,
            validity_period=validity_period,
            registered_delivery=registered_delivery,
            replace_if_present_flag=replace_if_present_flag,
            sm_default_msg_id=sm_default_msg_id,
            enquire_link_interval=enquire_link_interval,
            logger=logger,
            loglevel=loglevel,
            log_metadata=log_metadata,
            codec_class=codec_class,
            codec_errors_level=codec_errors_level,
            rateLimiter=rateLimiter,
            hook=hook,
            sequence_generator=sequence_generator,
            throttle_handler=throttle_handler,
            correlation_handler=correlation_handler,
            drain_duration=drain_duration,
            socket_timeout=socket_timeout,
        )

        self._PID = os.getpid()
        self.smsc_host = smsc_host
        self.smsc_port = smsc_port
        self.system_id = system_id
        self.password = password
        self.outboundqueue = outboundqueue

        if client_id is not None:
            self.client_id = client_id
        else:
            self.client_id = "".join(random.choices(string.ascii_uppercase + string.digits, k=17))

        self.system_type = system_type
        self.interface_version = interface_version
        self.addr_ton = addr_ton
        self.addr_npi = addr_npi
        self.address_range = address_range
        self.encoding = encoding

        if sequence_generator is not None:
            self.sequence_generator = sequence_generator
        else:
            self.sequence_generator = sequence.SimpleSequenceGenerator()

        self.max_sequence_number = 0x7FFFFFFF
        self.loglevel = loglevel.upper()

        if log_metadata is not None:
            self.log_metadata = log_metadata
        else:
            self.log_metadata = {}
        self.log_metadata.update(
            {
                "smsc_host": self.smsc_host,
                "system_id": system_id,
                "client_id": self.client_id,
                "pid": self._PID,
            }
        )

        self.codec_errors_level = codec_errors_level
        if codec_class is not None:
            self.codec_class = codec_class
        else:
            self.codec_class = nazcodec.SimpleNazCodec()

        self.service_type = service_type
        self.source_addr_ton = source_addr_ton
        self.source_addr_npi = source_addr_npi
        self.dest_addr_ton = dest_addr_ton
        self.dest_addr_npi = dest_addr_npi
        self.esm_class = esm_class
        self.protocol_id = protocol_id
        self.priority_flag = priority_flag
        self.schedule_delivery_time = schedule_delivery_time
        self.validity_period = validity_period
        self.registered_delivery = registered_delivery
        self.replace_if_present_flag = replace_if_present_flag
        self.sm_default_msg_id = sm_default_msg_id
        self.enquire_link_interval = enquire_link_interval

        # see section 5.1.2.1 of smpp ver 3.4 spec document
        self.command_ids = {
            SmppCommand.BIND_TRANSCEIVER: 0x00000009,
            SmppCommand.BIND_TRANSCEIVER_RESP: 0x80000009,
            SmppCommand.UNBIND: 0x00000006,
            SmppCommand.UNBIND_RESP: 0x80000006,
            SmppCommand.SUBMIT_SM: 0x00000004,
            SmppCommand.SUBMIT_SM_RESP: 0x80000004,
            SmppCommand.DELIVER_SM: 0x00000005,
            SmppCommand.DELIVER_SM_RESP: 0x80000005,
            SmppCommand.ENQUIRE_LINK: 0x00000015,
            SmppCommand.ENQUIRE_LINK_RESP: 0x80000015,
            SmppCommand.GENERIC_NACK: 0x80000000,
            # naz currently does not handle the following smpp commands.
            # open a github issue if you use naz and require support of a command in this list
            SmppCommand.BIND_RECEIVER_RESP: 0x80000001,
            SmppCommand.BIND_TRANSMITTER_RESP: 0x80000002,
            SmppCommand.QUERY_SM: 0x00000003,
            SmppCommand.QUERY_SM_RESP: 0x80000003,
            SmppCommand.REPLACE_SM: 0x00000007,
            SmppCommand.REPLACE_SM_RESP: 0x80000007,
            SmppCommand.CANCEL_SM: 0x00000008,
            SmppCommand.CANCEL_SM_RESP: 0x80000008,
            SmppCommand.SUBMIT_MULTI: 0x00000021,
            SmppCommand.SUBMIT_MULTI_RESP: 0x80000021,
            SmppCommand.OUTBIND: 0x0000000B,
            SmppCommand.ALERT_NOTIFICATION: 0x00000102,
            SmppCommand.DATA_SM: 0x00000103,
            SmppCommand.DATA_SM_RESP: 0x80000103,
            SmppCommand.RESERVED_A: 0x0000000A,
            SmppCommand.RESERVED_B: 0x8000000A,
            SmppCommand.RESERVED_C: 0x00000100,
            SmppCommand.RESERVED_D: 0x80000100,
            SmppCommand.RESERVED_E: 0x00000101,
            SmppCommand.RESERVED_F: 0x80000101,
            SmppCommand.RESERVED_G: 0x80000102,
            SmppCommand.RESERVED_LIST_A: [0x0000000C, 0x00000014],
            SmppCommand.RESERVED_LIST_B: [0x8000000B, 0x80000014],
            SmppCommand.RESERVED_LIST_C: [0x00000016, 0x00000020],
            SmppCommand.RESERVED_LIST_D: [0x80000016, 0x80000020],
            SmppCommand.RESERVED_LIST_E: [0x00000022, 0x000000FF],
            SmppCommand.RESERVED_LIST_F: [0x80000022, 0x800000FF],
            SmppCommand.RESERVED_LIST_G: [0x00010300, 0xFFFFFFFF],
            SmppCommand.RESERVED_LIST_H: [0x00010000, 0x000101FF],
            SmppCommand.RESERVED_LIST_I: [0x80010000, 0x800101FF],
            SmppCommand.RESERVED_FOR_SMPP_EXTENSION_A: [0x00000104, 0x0000FFFF],
            SmppCommand.RESERVED_FOR_SMPP_EXTENSION_B: [0x80000104, 0x8000FFFF],
            SmppCommand.RESERVED_FOR_SMSC_VENDOR_A: [0x00010200, 0x000102FF],
            SmppCommand.RESERVED_FOR_SMSC_VENDOR_B: [0x80010200, 0x800102FF],
        }

        self.data_coding = self._find_data_coding(self.encoding)

        self.reader: typing.Union[None, asyncio.streams.StreamReader] = None
        self.writer: typing.Union[None, asyncio.streams.StreamWriter] = None

        if logger is not None:
            self.logger = logger
        else:
            self.logger = log.SimpleLogger("naz.client")
        self.logger.bind(level=self.loglevel, log_metadata=self.log_metadata)
        self._sanity_check_logger()

        if rateLimiter is not None:
            self.rateLimiter = rateLimiter
        else:
            self.rateLimiter = ratelimiter.SimpleRateLimiter(logger=self.logger)

        if hook is not None:
            self.hook = hook
        else:
            self.hook = hooks.SimpleHook(logger=self.logger)

        if throttle_handler is not None:
            self.throttle_handler = throttle_handler
        else:
            self.throttle_handler = throttle.SimpleThrottleHandler(logger=self.logger)

        # class storing SMPP sequence_number and their corresponding log_id and/or hook_metadata
        # this will be used to track different pdu's and user generated log_id
        if correlation_handler is not None:
            self.correlation_handler = correlation_handler
        else:
            self.correlation_handler = correlater.SimpleCorrelater()

        # the messages that are published to a queue by either naz
        # or user application should be versioned.
        # This version will enable naz to be able to evolve in future;
        # eg a future version of naz could add/remove the number of required items in a message.
        # This is a bit similar to: http://docs.celeryproject.org/en/latest/internals/protocol.html
        self.naz_message_protocol_version = "1"

        self.current_session_state = SmppSessionState.CLOSED

        self.drain_duration = drain_duration
        self.socket_timeout = socket_timeout
        self.SHOULD_SHUT_DOWN: bool = False
        self.drain_lock: asyncio.Lock = asyncio.Lock()

    @staticmethod
    def _validate_client_args(
        smsc_host: str,
        smsc_port: int,
        system_id: str,
        password: str,
        outboundqueue: q.BaseOutboundQueue,
        client_id: typing.Union[None, str],
        system_type: str,
        addr_ton: int,
        addr_npi: int,
        address_range: str,
        encoding: str,
        interface_version: int,
        service_type: str,
        source_addr_ton: int,
        source_addr_npi: int,
        dest_addr_ton: int,
        dest_addr_npi: int,
        esm_class: int,
        protocol_id: int,
        priority_flag: int,
        schedule_delivery_time: str,
        validity_period: str,
        registered_delivery: int,
        replace_if_present_flag: int,
        sm_default_msg_id: int,
        enquire_link_interval: float,
        logger: typing.Union[None, log.BaseLogger],
        loglevel: str,
        log_metadata: typing.Union[None, dict],
        codec_class: typing.Union[None, nazcodec.BaseNazCodec],
        codec_errors_level: str,
        rateLimiter: typing.Union[None, ratelimiter.BaseRateLimiter],
        hook: typing.Union[None, hooks.BaseHook],
        sequence_generator: typing.Union[None, sequence.BaseSequenceGenerator],
        throttle_handler: typing.Union[None, throttle.BaseThrottleHandler],
        correlation_handler: typing.Union[None, correlater.BaseCorrelater],
        drain_duration: float,
        socket_timeout: float,
    ) -> None:
        """
        Checks that the arguments to `naz.Client` are okay.
        It raises an Exception that comprises of a list of Exceptions
        """
        errors: typing.List[ValueError] = []
        if not isinstance(smsc_host, str):
            errors.append(
                ValueError(
                    "`smsc_host` should be of type:: `str` You entered: {0}".format(type(smsc_host))
                )
            )
        if not isinstance(smsc_port, int):
            errors.append(
                ValueError(
                    "`smsc_port` should be of type:: `int` You entered: {0}".format(type(smsc_port))
                )
            )
        if not isinstance(system_id, str):
            errors.append(
                ValueError(
                    "`system_id` should be of type:: `str` You entered: {0}".format(type(system_id))
                )
            )
        if not isinstance(password, str):
            errors.append(
                ValueError(
                    "`password` should be of type:: `str` You entered: {0}".format(type(password))
                )
            )
        if not isinstance(outboundqueue, q.BaseOutboundQueue):
            errors.append(
                ValueError(
                    "`outboundqueue` should be of type:: `naz.q.BaseOutboundQueue` You entered: {0}".format(
                        type(outboundqueue)
                    )
                )
            )
        if not isinstance(client_id, (type(None), str)):
            errors.append(
                ValueError(
                    "`client_id` should be of type:: `None` or `str` You entered: {0}".format(
                        type(client_id)
                    )
                )
            )
        if not isinstance(system_type, str):
            errors.append(
                ValueError(
                    "`system_type` should be of type:: `str` You entered: {0}".format(
                        type(system_type)
                    )
                )
            )
        if not isinstance(addr_ton, int):
            errors.append(
                ValueError(
                    "`addr_ton` should be of type:: `int` You entered: {0}".format(type(addr_ton))
                )
            )
        if not isinstance(addr_npi, int):
            errors.append(
                ValueError(
                    "`addr_npi` should be of type:: `int` You entered: {0}".format(type(addr_npi))
                )
            )
        if not isinstance(address_range, str):
            errors.append(
                ValueError(
                    "`address_range` should be of type:: `str` You entered: {0}".format(
                        type(address_range)
                    )
                )
            )
        if not isinstance(encoding, str):
            errors.append(
                ValueError(
                    "`encoding` should be of type:: `str` You entered: {0}".format(type(encoding))
                )
            )
        if not isinstance(interface_version, int):
            errors.append(
                ValueError(
                    "`interface_version` should be of type:: `int` You entered: {0}".format(
                        type(interface_version)
                    )
                )
            )
        if not isinstance(service_type, str):
            errors.append(
                ValueError(
                    "`service_type` should be of type:: `str` You entered: {0}".format(
                        type(service_type)
                    )
                )
            )
        if not isinstance(source_addr_ton, int):
            errors.append(
                ValueError(
                    "`source_addr_ton` should be of type:: `int` You entered: {0}".format(
                        type(source_addr_ton)
                    )
                )
            )
        if not isinstance(source_addr_npi, int):
            errors.append(
                ValueError(
                    "`source_addr_npi` should be of type:: `int` You entered: {0}".format(
                        type(source_addr_npi)
                    )
                )
            )
        if not isinstance(dest_addr_ton, int):
            errors.append(
                ValueError(
                    "`dest_addr_ton` should be of type:: `int` You entered: {0}".format(
                        type(dest_addr_ton)
                    )
                )
            )
        if not isinstance(dest_addr_npi, int):
            errors.append(
                ValueError(
                    "`dest_addr_npi` should be of type:: `int` You entered: {0}".format(
                        type(dest_addr_npi)
                    )
                )
            )
        if not isinstance(esm_class, int):
            errors.append(
                ValueError(
                    "`esm_class` should be of type:: `int` You entered: {0}".format(type(esm_class))
                )
            )
        if not isinstance(protocol_id, int):
            errors.append(
                ValueError(
                    "`protocol_id` should be of type:: `int` You entered: {0}".format(
                        type(protocol_id)
                    )
                )
            )
        if not isinstance(priority_flag, int):
            errors.append(
                ValueError(
                    "`priority_flag` should be of type:: `int` You entered: {0}".format(
                        type(priority_flag)
                    )
                )
            )
        if not isinstance(schedule_delivery_time, str):
            errors.append(
                ValueError(
                    "`schedule_delivery_time` should be of type:: `str` You entered: {0}".format(
                        type(schedule_delivery_time)
                    )
                )
            )
        if not isinstance(validity_period, str):
            errors.append(
                ValueError(
                    "`validity_period` should be of type:: `str` You entered: {0}".format(
                        type(validity_period)
                    )
                )
            )
        if not isinstance(registered_delivery, int):
            errors.append(
                ValueError(
                    "`registered_delivery` should be of type:: `int` You entered: {0}".format(
                        type(registered_delivery)
                    )
                )
            )
        if not isinstance(replace_if_present_flag, int):
            errors.append(
                ValueError(
                    "`replace_if_present_flag` should be of type:: `int` You entered: {0}".format(
                        type(replace_if_present_flag)
                    )
                )
            )
        if not isinstance(sm_default_msg_id, int):
            errors.append(
                ValueError(
                    "`sm_default_msg_id` should be of type:: `int` You entered: {0}".format(
                        type(sm_default_msg_id)
                    )
                )
            )
        if not isinstance(enquire_link_interval, float):
            errors.append(
                ValueError(
                    "`enquire_link_interval` should be of type:: `float` You entered: {0}".format(
                        type(enquire_link_interval)
                    )
                )
            )
        if not isinstance(logger, (type(None), log.BaseLogger)):
            errors.append(
                ValueError(
                    "`logger` should be of type:: `None` or `naz.log.BaseLogger` You entered: {0}".format(
                        type(logger)
                    )
                )
            )
        if not isinstance(loglevel, str):
            errors.append(
                ValueError(
                    "`loglevel` should be of type:: `str` You entered: {0}".format(type(loglevel))
                )
            )
        if loglevel.upper() not in ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            errors.append(
                ValueError(
                    """`loglevel` should be one of; 'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'. You entered: {0}""".format(
                        loglevel
                    )
                )
            )
        if not isinstance(log_metadata, (type(None), dict)):
            errors.append(
                ValueError(
                    "`log_metadata` should be of type:: `None` or `dict` You entered: {0}".format(
                        type(log_metadata)
                    )
                )
            )
        if not isinstance(codec_class, (type(None), nazcodec.BaseNazCodec)):
            errors.append(
                ValueError(
                    "`codec_class` should be of type:: `None` or `naz.nazcodec.BaseNazCodec` You entered: {0}".format(
                        type(codec_class)
                    )
                )
            )
        if not isinstance(codec_errors_level, str):
            errors.append(
                ValueError(
                    "`codec_errors_level` should be of type:: `str` You entered: {0}".format(
                        type(codec_errors_level)
                    )
                )
            )
        if not isinstance(rateLimiter, (type(None), ratelimiter.BaseRateLimiter)):
            errors.append(
                ValueError(
                    "`rateLimiter` should be of type:: `None` or `naz.ratelimiter.BaseRateLimiter` You entered: {0}".format(
                        type(rateLimiter)
                    )
                )
            )
        if not isinstance(hook, (type(None), hooks.BaseHook)):
            errors.append(
                ValueError(
                    "`hook` should be of type:: `None` or `naz.hooks.BaseHook` You entered: {0}".format(
                        type(hook)
                    )
                )
            )
        if not isinstance(sequence_generator, (type(None), sequence.BaseSequenceGenerator)):
            errors.append(
                ValueError(
                    "`sequence_generator` should be of type:: `None` or `naz.sequence.BaseSequenceGenerator` You entered: {0}".format(
                        type(sequence_generator)
                    )
                )
            )
        if not isinstance(throttle_handler, (type(None), throttle.BaseThrottleHandler)):
            errors.append(
                ValueError(
                    "`throttle_handler` should be of type:: `None` or `naz.throttle.BaseThrottleHandler` You entered: {0}".format(
                        type(throttle_handler)
                    )
                )
            )
        if not isinstance(correlation_handler, (type(None), correlater.BaseCorrelater)):
            errors.append(
                ValueError(
                    "`correlation_handler` should be of type:: `None` or `naz.correlater.BaseCorrelater` You entered: {0}".format(
                        type(correlation_handler)
                    )
                )
            )
        if not isinstance(drain_duration, float):
            errors.append(
                ValueError(
                    "`drain_duration` should be of type:: `float` You entered: {0}".format(
                        type(drain_duration)
                    )
                )
            )
        if not isinstance(socket_timeout, float):
            errors.append(
                ValueError(
                    "`socket_timeout` should be of type:: `float` You entered: {0}".format(
                        type(socket_timeout)
                    )
                )
            )
        if len(errors):
            raise NazClientError(errors)

    def _sanity_check_logger(self):
        """
        called when instantiating the Client just to make sure the supplied
        logger can log.
        """
        try:
            self.logger.log(logging.DEBUG, {"event": "sanity_check_logger"})
        except Exception as e:
            raise e

    def _log(self, level, log_data):
        # if the supplied logger is unable to log; we move on
        try:
            self.logger.log(level, log_data)
        except Exception:
            pass

    @staticmethod
    def _find_data_coding(encoding):
        for key, val in SmppDataCoding.__dict__.items():
            if not key.startswith("__"):
                if encoding == val.code:
                    return val.value
        raise ValueError("That encoding:{0} is not recognised.".format(encoding))

    def _search_by_command_id_code(self, command_id_code: int) -> typing.Union[None, str]:
        for key, val in self.command_ids.items():
            if isinstance(val, list):
                __range = range(val[0], val[1] + 1)
                if command_id_code in __range:
                    return key
            else:
                if val == command_id_code:
                    return key
        return None

    @staticmethod
    def _search_by_command_status_value(
        command_status_value: int
    ) -> typing.Union[None, CommandStatus]:
        # TODO: find a cheaper(better) way of doing this
        for key, val in SmppCommandStatus.__dict__.items():
            if not key.startswith("__"):
                if isinstance(val.value, list):
                    __range = range(val.value[0], val.value[1] + 1)
                    if command_status_value in __range:
                        return val
                if command_status_value == val.value:
                    return val
        return None

    @staticmethod
    def _retry_after(current_retries):
        """
        retries will happen in this sequence;
        1min, 2min, 4min, 8min, 16min, 32min, 16min, 16min, 16min ...
        """
        # TODO:
        # 1. give users ability to bring their own retry algorithms.
        # 2. add jitter
        if current_retries < 0:
            current_retries = 0

        if current_retries >= 6:
            return 60 * 16  # 16 minutes
        else:
            return 60 * (1 * (2 ** current_retries))

    def _msg_to_log(self, msg: bytes) -> str:
        """
        returns decoded string from bytes with any password removed.
        the returned string is safe to log.
        """
        log_msg = ""
        try:
            log_msg = self.codec_class.decode(msg, self.encoding, self.codec_errors_level)
            if self.password in log_msg:
                # do not log password, redact it from logs.
                log_msg = log_msg.replace(self.password, "{REDACTED}")
        except (UnicodeDecodeError, UnicodeError):
            log_msg = str(msg)
        except Exception:
            pass
        return log_msg

    async def connect(
        self, log_id: str = ""
    ) -> typing.Tuple[asyncio.streams.StreamReader, asyncio.streams.StreamWriter]:
        """
        make a network connection to SMSC server.
        """
        if log_id == "":
            log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
        self._log(logging.INFO, {"event": "naz.Client.connect", "stage": "start", "log_id": log_id})
        reader, writer = await asyncio.open_connection(self.smsc_host, self.smsc_port)
        self.reader = reader
        self.writer = writer
        sock = self.writer.get_extra_info("socket")
        sock.settimeout(self.socket_timeout)
        # A socket object can be in one of three modes: blocking, non-blocking, or timeout.
        # At the OS level, sockets in timeout mode are internally set in non-blocking mode.
        # https://docs.python.org/3.6/library/socket.html#notes-on-socket-timeouts

        self._log(logging.INFO, {"event": "naz.Client.connect", "stage": "end", "log_id": log_id})
        self.current_session_state = SmppSessionState.OPEN
        return reader, writer

    async def tranceiver_bind(self, log_id: str = "") -> None:
        """
        send a BIND_RECEIVER pdu to SMSC.
        """
        smpp_command = SmppCommand.BIND_TRANSCEIVER
        if log_id == "":
            log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.tranceiver_bind",
                "stage": "start",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )
        # body
        body = b""
        body = (
            body
            + self.codec_class.encode(self.system_id, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + self.codec_class.encode(self.password, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + self.codec_class.encode(self.system_type, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + struct.pack(">I", self.interface_version)
            + struct.pack(">I", self.addr_ton)
            + struct.pack(">I", self.addr_npi)
            + self.codec_class.encode(self.address_range, self.encoding, self.codec_errors_level)
            + chr(0).encode()
        )

        # header
        command_length = 16 + len(body)  # 16 is for headers
        command_id = self.command_ids[smpp_command]
        # the status for success see section 5.1.3
        command_status = SmppCommandStatus.ESME_ROK.value
        try:
            sequence_number = self.sequence_generator.next_sequence()
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.tranceiver_bind",
                    "stage": "end",
                    "error": str(e),
                    "smpp_command": smpp_command,
                },
            )

        if sequence_number > self.max_sequence_number:
            # prevent third party sequence_generators from ruining our party
            raise ValueError(
                "the sequence_number: {0} is greater than the max: {1} allowed by SMPP spec.".format(
                    sequence_number, self.max_sequence_number
                )
            )

        # associate sequence_number with log_id.
        # this will enable us to also associate responses and thus enhancing traceability of all workflows
        try:
            await self.correlation_handler.put(
                smpp_command=smpp_command,
                sequence_number=sequence_number,
                log_id=log_id,
                hook_metadata="",
            ),
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.tranceiver_bind",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "correlater put error",
                    "error": str(e),
                },
            )

        header = struct.pack(">IIII", command_length, command_id, command_status, sequence_number)
        full_pdu = header + body
        await self.send_data(smpp_command=smpp_command, msg=full_pdu, log_id=log_id)
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.tranceiver_bind",
                "stage": "end",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

    async def enquire_link(self, TESTING: bool = False) -> typing.Union[None, bytes]:
        """
        send an ENQUIRE_LINK pdu to SMSC.

        Parameters:
            TESTING: indicates whether this method is been called while running tests.
        """
        # sleep during startup so that `naz` can have had time to connect & bind
        # we rely on `enquire_link` to kick on `re_establish_conn_bind`
        while self.current_session_state != SmppSessionState.BOUND_TRX:
            retry_after = self.socket_timeout / 5
            self._log(
                logging.DEBUG,
                {
                    "event": "naz.Client.enquire_link",
                    "stage": "start",
                    "current_session_state": self.current_session_state,
                    "state": "awaiting naz to change session state to `BOUND_TRX`. sleeping for {0:.2f} seconds".format(
                        retry_after
                    ),
                },
            )
            await asyncio.sleep(retry_after)
            if TESTING:
                return None

        smpp_command = SmppCommand.ENQUIRE_LINK
        while True:
            log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
            self._log(
                logging.DEBUG,
                {
                    "event": "naz.Client.enquire_link",
                    "stage": "start",
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )
            if self.SHOULD_SHUT_DOWN:
                self._log(
                    logging.DEBUG,
                    {
                        "event": "naz.Client.enquire_link",
                        "stage": "end",
                        "state": "client is shutdown",
                    },
                )
                return None

            # body
            body = b""

            # header
            command_length = 16 + len(body)  # 16 is for headers
            command_id = self.command_ids[smpp_command]
            command_status = 0x00000000  # not used for `enquire_link`
            try:
                sequence_number = self.sequence_generator.next_sequence()
            except Exception as e:
                self._log(
                    logging.ERROR,
                    {
                        "event": "naz.Client.enquire_link",
                        "stage": "end",
                        "error": str(e),
                        "log_id": log_id,
                        "smpp_command": smpp_command,
                    },
                )
            if sequence_number > self.max_sequence_number:
                # prevent third party sequence_generators from ruining our party
                raise ValueError(
                    "the sequence_number: {0} is greater than the max: {1} allowed by SMPP spec.".format(
                        sequence_number, self.max_sequence_number
                    )
                )

            try:
                await self.correlation_handler.put(
                    smpp_command=smpp_command,
                    sequence_number=sequence_number,
                    log_id=log_id,
                    hook_metadata="",
                )
            except Exception as e:
                self._log(
                    logging.ERROR,
                    {
                        "event": "naz.Client.enquire_link",
                        "stage": "end",
                        "smpp_command": smpp_command,
                        "log_id": log_id,
                        "state": "correlater put error",
                        "error": str(e),
                    },
                )

            header = struct.pack(
                ">IIII", command_length, command_id, command_status, sequence_number
            )
            full_pdu = header + body
            # dont queue enquire_link in SimpleOutboundQueue since we dont want it to be behind 10k msgs etc
            await self.send_data(smpp_command=smpp_command, msg=full_pdu, log_id=log_id)
            self._log(
                logging.DEBUG,
                {
                    "event": "naz.Client.enquire_link",
                    "stage": "end",
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )
            if TESTING:
                return full_pdu
            await asyncio.sleep(self.enquire_link_interval)

    async def enquire_link_resp(self, sequence_number: int) -> None:
        """
        send an ENQUIRE_LINK_RESP pdu to SMSC.

        Parameters:
            sequence_number: SMPP sequence_number
        """
        smpp_command = SmppCommand.ENQUIRE_LINK_RESP
        log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
        self._log(
            logging.DEBUG,
            {
                "event": "naz.Client.enquire_link_resp",
                "stage": "start",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

        # body
        body = b""

        # header
        command_length = 16 + len(body)  # 16 is for headers
        command_id = self.command_ids[smpp_command]
        command_status = SmppCommandStatus.ESME_ROK.value
        sequence_number = sequence_number
        header = struct.pack(">IIII", command_length, command_id, command_status, sequence_number)

        full_pdu = header + body
        item_to_enqueue = {
            "version": self.naz_message_protocol_version,
            "log_id": log_id,
            "pdu": full_pdu,
            "smpp_command": smpp_command,
        }
        try:
            await self.outboundqueue.enqueue(item_to_enqueue)
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.enquire_link_resp",
                    "stage": "end",
                    "error": str(e),
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )
        self._log(
            logging.DEBUG,
            {
                "event": "naz.Client.enquire_link_resp",
                "stage": "end",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

    async def unbind_resp(self, sequence_number: int) -> None:
        """
        send an UNBIND_RESP pdu to SMSC.

        Parameters:
            sequence_number: SMPP sequence_number
        """
        smpp_command = SmppCommand.UNBIND_RESP
        log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.unbind_resp",
                "stage": "start",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

        # body
        body = b""

        # header
        command_length = 16 + len(body)  # 16 is for headers
        command_id = self.command_ids[smpp_command]
        command_status = SmppCommandStatus.ESME_ROK.value
        sequence_number = sequence_number
        header = struct.pack(">IIII", command_length, command_id, command_status, sequence_number)

        full_pdu = header + body
        # dont queue unbind_resp in SimpleOutboundQueue since we dont want it to be behind 10k msgs etc
        await self.send_data(smpp_command=smpp_command, msg=full_pdu, log_id=log_id)
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.unbind_resp",
                "stage": "end",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

    async def deliver_sm_resp(self, sequence_number: int) -> None:
        """
        send a DELIVER_SM_RESP pdu to SMSC.

        Parameters:
            sequence_number: SMPP sequence_number
        """
        smpp_command = SmppCommand.DELIVER_SM_RESP
        log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.deliver_sm_resp",
                "stage": "start",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )
        # body
        body = b""
        message_id = ""
        body = (
            body
            + self.codec_class.encode(message_id, self.encoding, self.codec_errors_level)
            + chr(0).encode()
        )

        # header
        command_length = 16 + len(body)  # 16 is for headers
        command_id = self.command_ids[smpp_command]
        command_status = SmppCommandStatus.ESME_ROK.value
        sequence_number = sequence_number
        header = struct.pack(">IIII", command_length, command_id, command_status, sequence_number)

        full_pdu = header + body
        item_to_enqueue = {
            "version": self.naz_message_protocol_version,
            "log_id": log_id,
            "pdu": full_pdu,
            "smpp_command": smpp_command,
        }
        try:
            await self.outboundqueue.enqueue(item_to_enqueue)
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.deliver_sm_resp",
                    "stage": "end",
                    "error": str(e),
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )

        self._log(
            logging.INFO,
            {
                "event": "naz.Client.deliver_sm_resp",
                "stage": "end",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

    # this method just enqueues a submit_sm msg to queue
    async def submit_sm(
        self, short_message: str, log_id: str, source_addr: str, destination_addr: str
    ) -> None:
        """
        enqueues a SUBMIT_SM pdu to :attr:`outboundqueue <Client.outboundqueue>`
        That PDU will later on be sent to SMSC.

        Parameters:
            short_message: message to send to SMSC
            log_id: a unique identify of this request
            source_addr: the identifier(eg msisdn) of the message sender
            destination_addr: the identifier(eg msisdn) of the message recipient
        """
        # HEADER::
        # submit_sm has the following pdu header:
        # command_length, int, 4octet
        # command_id, int, 4octet. `submit_sm`
        # command_status, int, 4octet. Not used. Set to NULL
        # sequence_number, int, 4octet.  The associated submit_sm_resp PDU will echo this sequence number.

        # BODY::
        # submit_sm has the following pdu body. NB: They SHOULD be put in the body in the ORDER presented here.
        # service_type, c-octet str, max 6octet. eg NULL, "USSD", "CMT" etc
        # source_addr_ton, int , 1octet,
        # source_addr_npi, int, 1octet
        # source_addr, c-octet str, max 21octet. eg; This is usually the senders phone Number
        # dest_addr_ton, int, 1octet
        # dest_addr_npi, int, 1octet
        # destination_addr,  C-Octet String, max 21 octet. eg; This is usually the recipients phone Number
        # esm_class, int, 1octet
        # protocol_id, int, 1octet
        # priority_flag, int, 1octet
        # schedule_delivery_time, c-octet str, 1 or 17 octets. NULL for immediate message delivery.
        # validity_period, c-octet str, 1 or 17 octets.  NULL for SMSC default.
        # registered_delivery, int, 1octet
        # replace_if_present_flag, int, 1octet
        # data_coding, int, 1octet. Defines the encoding scheme of the short message user data. Bits 7 6 5 4 3 2 1 0
        # sm_default_msg_id, int, 1octet. SMSC index of a pre-defined(`canned`) message.  If not using an SMSC canned message, set to NULL
        # sm_length, int, 1octet. Length in octets of the `short_message`.
        # short_message, Octet-String(NOT c-octet str), 0-254 octets.
        # NB: 1. Applications which need to send messages longer than 254 octets should use the `message_payload` optional parameter.
        #        In this case the `sm_length` field should be set to zero
        #        u cant use both `short_message` and `message_payload`
        #     2. Octet String - A series of octets, not necessarily NULL terminated.

        smpp_command = SmppCommand.SUBMIT_SM
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.submit_sm",
                "stage": "start",
                "log_id": log_id,
                "short_message": short_message,
                "source_addr": source_addr,
                "destination_addr": destination_addr,
                "smpp_command": smpp_command,
            },
        )
        item_to_enqueue = {
            "version": self.naz_message_protocol_version,
            "smpp_command": smpp_command,
            "short_message": short_message,
            "log_id": log_id,
            "source_addr": source_addr,
            "destination_addr": destination_addr,
        }
        try:
            await self.outboundqueue.enqueue(item_to_enqueue)
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.submit_sm",
                    "stage": "end",
                    "error": str(e),
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.submit_sm",
                "stage": "end",
                "log_id": log_id,
                "short_message": short_message,
                "source_addr": source_addr,
                "destination_addr": destination_addr,
                "smpp_command": smpp_command,
            },
        )

    async def _build_submit_sm_pdu(
        self, short_message, log_id, hook_metadata, source_addr, destination_addr
    ) -> bytes:
        """
        builds a SUBMIT_SM pdu.

        Parameters:
            short_message: message to send to SMSC
            log_id: a unique identify of this request
            hook_metadata: additional metadata that you would like to be passed on to hooks
            source_addr: the identifier(eg msisdn) of the message sender
            destination_addr: the identifier(eg msisdn) of the message recipient
        """
        smpp_command = SmppCommand.SUBMIT_SM
        self._log(
            logging.DEBUG,
            {
                "event": "naz.Client._build_submit_sm_pdu",
                "stage": "start",
                "log_id": log_id,
                "short_message": short_message,
                "source_addr": source_addr,
                "destination_addr": destination_addr,
                "smpp_command": smpp_command,
            },
        )
        encoded_short_message = self.codec_class.encode(
            short_message, self.encoding, self.codec_errors_level
        )
        sm_length = len(encoded_short_message)

        # body
        body = b""
        body = (
            body
            + self.codec_class.encode(self.service_type, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + struct.pack(">B", self.source_addr_ton)
            + struct.pack(">B", self.source_addr_npi)
            + self.codec_class.encode(source_addr, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + struct.pack(">B", self.dest_addr_ton)
            + struct.pack(">B", self.dest_addr_npi)
            + self.codec_class.encode(destination_addr, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + struct.pack(">B", self.esm_class)
            + struct.pack(">B", self.protocol_id)
            + struct.pack(">B", self.priority_flag)
            + self.codec_class.encode(
                self.schedule_delivery_time, self.encoding, self.codec_errors_level
            )
            + chr(0).encode()
            + self.codec_class.encode(self.validity_period, self.encoding, self.codec_errors_level)
            + chr(0).encode()
            + struct.pack(">B", self.registered_delivery)
            + struct.pack(">B", self.replace_if_present_flag)
            + struct.pack(">B", self.data_coding)
            + struct.pack(">B", self.sm_default_msg_id)
            + struct.pack(">B", sm_length)
            + self.codec_class.encode(short_message, self.encoding, self.codec_errors_level)
        )

        # header
        command_length = 16 + len(body)  # 16 is for headers
        command_id = self.command_ids[smpp_command]
        # the status for success see section 5.1.3
        command_status = 0x00000000  # not used for `submit_sm`
        try:
            sequence_number = self.sequence_generator.next_sequence()
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client._build_submit_sm_pdu",
                    "stage": "end",
                    "error": str(e),
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )
        if sequence_number > self.max_sequence_number:
            # prevent third party sequence_generators from ruining our party
            raise ValueError(
                "the sequence_number: {0} is greater than the max: {1} allowed by SMPP spec.".format(
                    sequence_number, self.max_sequence_number
                )
            )

        try:
            await self.correlation_handler.put(
                smpp_command=smpp_command,
                sequence_number=sequence_number,
                log_id=log_id,
                hook_metadata=hook_metadata,
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client._build_submit_sm_pdu",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "correlater put error",
                    "error": str(e),
                },
            )

        header = struct.pack(">IIII", command_length, command_id, command_status, sequence_number)
        full_pdu = header + body
        self._log(
            logging.DEBUG,
            {
                "event": "naz.Client._build_submit_sm_pdu",
                "stage": "end",
                "log_id": log_id,
                "short_message": short_message,
                "source_addr": source_addr,
                "destination_addr": destination_addr,
                "smpp_command": smpp_command,
            },
        )
        return full_pdu

    async def re_establish_conn_bind(
        self, smpp_command: str, log_id: str, TESTING: bool = False
    ) -> None:
        """
        Called if connection is lost. It reconnects & rebinds to SMSC.

        Parameters:
            TESTING: indicates whether this method is been called while running tests.
        """
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.re_establish_conn_bind",
                "stage": "start",
                "smpp_command": smpp_command,
                "log_id": log_id,
                "connection_lost": self.writer.transport.is_closing() if self.writer else True,
            },
        )
        if self.SHOULD_SHUT_DOWN:
            self._log(
                logging.DEBUG,
                {
                    "event": "naz.Client.re_establish_conn_bind",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "cleanly shutting down client.",
                },
            )
            return None

        try:
            # 1. re-connect
            # 2. re-bind
            await self.connect(log_id=log_id)
            await self.tranceiver_bind(log_id=log_id)
        except (
            ConnectionError,
            TimeoutError,
            asyncio.TimeoutError,
            socket.error,
            socket.herror,
            socket.gaierror,
            socket.timeout,
        ) as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.re_establish_conn_bind",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "unable to re-connect & re-bind to SMSC",
                    "error": str(e),
                },
            )
        if TESTING:
            # offer escape hatch for tests to come out of endless loop
            return None

    async def send_data(
        self, smpp_command: str, msg: bytes, log_id: str, hook_metadata: str = ""
    ) -> None:
        """
        Sends PDU's to SMSC through/over a network connection(down the wire).
        This method does not block; it buffers the data and arranges for it to be sent out asynchronously.
        It also accts as a flow control method that interacts with the IO write buffer.

        Parameters:
            smpp_command: type of PDU been sent. eg bind_transceiver
            msg: PDU to be sent to SMSC over the network connection.
            log_id: a unique identify of this request
            hook_metadata: additional metadata that you would like to be passed on to hooks
        """
        # todo: look at `set_write_buffer_limits` and `get_write_buffer_limits` methods
        # print("get_write_buffer_limits:", writer.transport.get_write_buffer_limits())

        if isinstance(msg, str):
            msg = self.codec_class.encode(msg, self.encoding, self.codec_errors_level)
        log_msg = self._msg_to_log(msg=msg)
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.send_data",
                "stage": "start",
                "smpp_command": smpp_command,
                "log_id": log_id,
                "msg": log_msg,
                "connection_lost": self.writer.transport.is_closing() if self.writer else True,
            },
        )

        # check session state to see if we can send messages.
        # see section 2.3 of SMPP spec document v3.4
        if self.current_session_state == SmppSessionState.CLOSED:
            error_msg = "smpp_command `{0}` cannot be sent to SMSC when the client session state is `{1}`".format(
                smpp_command, self.current_session_state
            )
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.send_data",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "msg": log_msg,
                    "current_session_state": self.current_session_state,
                    "error": error_msg,
                },
            )
            return None
        elif (
            self.current_session_state == SmppSessionState.OPEN
            and smpp_command == SmppCommand.ENQUIRE_LINK
        ):
            # If the connection to  SMSC is broken, we need to call `Client.re_establish_conn_bind`.
            # That method is only called by `Client.send_data`. Thus someone has to call `Client.send_data` even
            # when the connection is broken in order for it to call `re_establish_conn_bind` and restore connection
            # That someone is `enquire_link`. This is why in this block we have `enquire_link` and logging it at DEBUG level
            # and we do not return
            error_msg = "smpp_command `{0}` cannot be sent to SMSC when the client session state is `{1}`".format(
                smpp_command, self.current_session_state
            )
            self._log(
                logging.DEBUG,
                {
                    "event": "naz.Client.send_data",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "msg": log_msg,
                    "current_session_state": self.current_session_state,
                    "error": error_msg,
                },
            )
            # do not raise or return
        elif self.current_session_state == SmppSessionState.OPEN and smpp_command not in [
            SmppCommand.BIND_TRANSMITTER,
            SmppCommand.BIND_RECEIVER,
            SmppCommand.BIND_TRANSCEIVER,
        ]:
            # only the smpp_command's listed above are allowed by SMPP spec to be sent
            # if current_session_state == SmppSessionState.OPEN
            error_msg = "smpp_command `{0}` cannot be sent to SMSC when the client session state is `{1}`".format(
                smpp_command, self.current_session_state
            )
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.send_data",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "msg": log_msg,
                    "current_session_state": self.current_session_state,
                    "error": error_msg,
                },
            )
            # do not raise, we do not want naz-cli to exit
            return None

        if (self.writer is None) or self.writer.transport.is_closing():
            await self.re_establish_conn_bind(smpp_command=smpp_command, log_id=log_id)

        try:
            # call user's hook for requests
            await self.hook.request(
                smpp_command=smpp_command, log_id=log_id, hook_metadata=hook_metadata
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.send_data",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "request hook error",
                    "error": str(e),
                },
            )

        try:
            if typing.TYPE_CHECKING:
                # make mypy happy; https://github.com/python/mypy/issues/4805
                assert isinstance(self.writer, asyncio.streams.StreamWriter)

            # We use writer.drain() which is a flow control method that interacts with the IO write buffer.
            # When the size of the buffer reaches the high watermark,
            # drain blocks until the size of the buffer is drained down to the low watermark and writing can be resumed.
            # When there is nothing to wait for, the drain() returns immediately.
            # ref: https://docs.python.org/3/library/asyncio-stream.html#asyncio.StreamWriter.drain
            self.writer.write(msg)
            async with self.drain_lock:
                # see: https://github.com/komuw/naz/issues/114
                await self.writer.drain()
            if smpp_command == SmppCommand.BIND_TRANSCEIVER:
                # if we have successfully sent a bind_transceiver request, we can set session state to `BOUND_TRX`
                # Ideally, you should only set state to `BOUND_TRX` once SMSC sends back a successful `BIND_TRANSCEIVER_RESP`
                # However, an SMSC may fail to do so. This is especially true when sending `re_establish_conn_bind`
                # hack!! bad!!
                # TODO: fix this
                self.current_session_state = SmppSessionState.BOUND_TRX
        except (
            ConnectionError,
            TimeoutError,
            asyncio.TimeoutError,
            socket.error,
            socket.herror,
            socket.gaierror,
            socket.timeout,
        ) as e:
            # https://docs.python.org/3.6/library/socket.html#exceptions
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.send_data",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "unable to write to SMSC",
                    "error": str(e),
                },
            )

        self._log(
            logging.INFO,
            {
                "event": "naz.Client.send_data",
                "stage": "end",
                "smpp_command": smpp_command,
                "log_id": log_id,
                "msg": log_msg,
            },
        )

    async def dequeue_messages(
        self, TESTING: bool = False
    ) -> typing.Union[str, typing.Dict[typing.Any, typing.Any]]:
        """
        In a loop; dequeues items from the :attr:`outboundqueue <Client.outboundqueue>` and sends them to SMSC.

        Parameters:
            TESTING: indicates whether this method is been called while running tests.
        """
        dequeue_retry_count = 0
        while True:
            self._log(logging.INFO, {"event": "naz.Client.dequeue_messages", "stage": "start"})
            if self.SHOULD_SHUT_DOWN:
                self._log(
                    logging.INFO,
                    {
                        "event": "naz.Client.dequeue_messages",
                        "stage": "end",
                        "state": "cleanly shutting down client.",
                    },
                )
                return {"shutdown": "shutdown"}

            while self.current_session_state != SmppSessionState.BOUND_TRX:
                # If the connection to SMSC is broken, there's no need to try and send messages
                # sleep and wait for `Client.re_establish_conn_bind` to do its thing.
                # this same thing cannot be done for `enquire_link` since we rely on it to kick on `re_establish_conn_bind`
                retry_after = self.socket_timeout
                self._log(
                    logging.INFO,
                    {
                        "event": "naz.Client.dequeue_messages",
                        "stage": "start",
                        "current_session_state": self.current_session_state,
                        "state": "awaiting naz to change session state to `BOUND_TRX`. sleeping for {0:.2f} seconds".format(
                            retry_after
                        ),
                    },
                )
                await asyncio.sleep(retry_after)
                if TESTING:
                    return {"state": "awaiting naz to change session state to `BOUND_TRX`"}

            # TODO: there are so many try-except classes in this func.
            # do something about that.
            try:
                # check with throttle handler
                send_request = await self.throttle_handler.allow_request()
            except Exception as e:
                self._log(
                    logging.ERROR,
                    {
                        "event": "naz.Client.dequeue_messages",
                        "stage": "end",
                        "state": "dequeue_messages error",
                        "error": str(e),
                    },
                )
                continue
            if send_request:
                try:
                    # rate limit ourselves
                    await self.rateLimiter.limit()
                except Exception as e:
                    self._log(
                        logging.ERROR,
                        {
                            "event": "naz.Client.dequeue_messages",
                            "stage": "end",
                            "state": "dequeue_messages error",
                            "error": str(e),
                        },
                    )

                try:
                    item_to_dequeue = await self.outboundqueue.dequeue()
                except Exception as e:
                    dequeue_retry_count += 1
                    poll_queue_interval = self._retry_after(dequeue_retry_count)
                    self._log(
                        logging.ERROR,
                        {
                            "event": "naz.Client.dequeue_messages",
                            "stage": "end",
                            "state": "dequeue_messages error. sleeping for {0:.2f} seconds".format(
                                poll_queue_interval
                            ),
                            "dequeue_retry_count": dequeue_retry_count,
                            "error": str(e),
                        },
                    )
                    if self.SHOULD_SHUT_DOWN:
                        return {"shutdown": "shutdown"}
                    if TESTING:
                        # offer escape hatch for tests to come out of endless loop
                        return {"broker_error": "broker_error"}
                    await asyncio.sleep(poll_queue_interval)
                    continue

                # we didn't fail to dequeue a message
                dequeue_retry_count = 0
                try:
                    log_id = item_to_dequeue["log_id"]
                    item_to_dequeue["version"]  # version is a required field
                    smpp_command = item_to_dequeue["smpp_command"]
                    hook_metadata = item_to_dequeue.get("hook_metadata", "")
                    if smpp_command == SmppCommand.SUBMIT_SM:
                        short_message = item_to_dequeue["short_message"]
                        source_addr = item_to_dequeue["source_addr"]
                        destination_addr = item_to_dequeue["destination_addr"]
                        full_pdu = await self._build_submit_sm_pdu(
                            short_message, log_id, hook_metadata, source_addr, destination_addr
                        )
                    else:
                        full_pdu = item_to_dequeue["pdu"]
                except KeyError as e:
                    e = KeyError(
                        "enqueued message/object is missing required field: {}".format(str(e))
                    )
                    self._log(
                        logging.ERROR,
                        {
                            "event": "naz.Client.dequeue_messages",
                            "stage": "end",
                            "state": "dequeue_messages error",
                            "error": str(e),
                        },
                    )
                    continue

                await self.send_data(
                    smpp_command=smpp_command,
                    msg=full_pdu,
                    log_id=log_id,
                    hook_metadata=hook_metadata,
                )
                self._log(
                    logging.INFO,
                    {
                        "event": "naz.Client.dequeue_messages",
                        "stage": "end",
                        "log_id": log_id,
                        "smpp_command": smpp_command,
                        "send_request": send_request,
                    },
                )
                if TESTING:
                    # offer escape hatch for tests to come out of endless loop
                    return item_to_dequeue
            else:
                # throttle_handler didn't allow us to send request.
                self._log(
                    logging.INFO,
                    {
                        "event": "naz.Client.dequeue_messages",
                        "stage": "end",
                        "send_request": send_request,
                    },
                )
                try:
                    await asyncio.sleep(await self.throttle_handler.throttle_delay())
                except Exception as e:
                    self._log(
                        logging.ERROR,
                        {
                            "event": "naz.Client.dequeue_messages",
                            "stage": "end",
                            "state": "dequeue_messages error",
                            "error": str(e),
                        },
                    )
                    continue
                if TESTING:
                    # offer escape hatch for tests to come out of endless loop
                    return {"throttle_handler_denied_request": "throttle_handler_denied_request"}
                continue

    async def receive_data(self, TESTING: bool = False) -> typing.Union[None, bytes]:
        """
        In a loop; read bytes from the network connected to SMSC and hand them over to the :func:`throparserttled <Client._parse_response_pdu>`.

        Parameters:
            TESTING: indicates whether this method is been called while running tests.
        """
        retry_count = 0
        while True:
            self._log(logging.INFO, {"event": "naz.Client.receive_data", "stage": "start"})
            if self.SHOULD_SHUT_DOWN:
                self._log(
                    logging.INFO,
                    {
                        "event": "naz.Client.receive_data",
                        "stage": "end",
                        "state": "cleanly shutting down client.",
                    },
                )
                return None

            command_length_header_data = b""
            try:
                if typing.TYPE_CHECKING:
                    # make mypy happy; https://github.com/python/mypy/issues/4805
                    assert isinstance(self.reader, asyncio.streams.StreamReader)

                # todo: look at `pause_reading` and `resume_reading` methods
                # `client.reader` and `client.writer` should not have timeouts since they are non-blocking
                # https://github.com/komuw/naz/issues/116
                command_length_header_data = await self.reader.read(4)
            except (
                ConnectionError,
                TimeoutError,
                asyncio.TimeoutError,
                socket.error,
                socket.herror,
                socket.gaierror,
                socket.timeout,
            ) as e:
                self._log(
                    logging.ERROR,
                    {
                        "event": "naz.Client.receive_data",
                        "stage": "end",
                        "state": "unable to read from SMSC",
                        "error": str(e),
                    },
                )

            if command_length_header_data == b"":
                retry_count += 1
                poll_read_interval = self._retry_after(retry_count)
                self._log(
                    logging.INFO,
                    {
                        "event": "naz.Client.receive_data",
                        "stage": "start",
                        "state": "no data received from SMSC. sleeping for {0:.2f} seconds".format(
                            poll_read_interval
                        ),
                        "retry_count": retry_count,
                    },
                )
                if self.SHOULD_SHUT_DOWN:
                    return None
                await asyncio.sleep(poll_read_interval)
                continue
            else:
                # we didn't fail to read from SMSC
                retry_count = 0

            total_pdu_length = struct.unpack(">I", command_length_header_data)[0]
            MSGLEN = total_pdu_length - 4
            chunks = []
            bytes_recd = 0
            while bytes_recd < MSGLEN:
                chunk = b""
                try:
                    if typing.TYPE_CHECKING:
                        # make mypy happy; https://github.com/python/mypy/issues/4805
                        assert isinstance(self.reader, asyncio.streams.StreamReader)

                    chunk = await self.reader.read(min(MSGLEN - bytes_recd, 2048))
                    if chunk == b"":
                        # TODO: maybe we also need todo; `self.writer=None`
                        # so that the `re_establish_conn_bind` mechanism can kick in.
                        raise ConnectionError("socket connection broken")
                except (
                    ConnectionError,
                    TimeoutError,
                    asyncio.TimeoutError,
                    socket.error,
                    socket.herror,
                    socket.gaierror,
                    socket.timeout,
                ) as e:
                    self._log(
                        logging.ERROR,
                        {
                            "event": "naz.Client.receive_data",
                            "stage": "end",
                            "state": "unable to read from SMSC",
                            "error": str(e),
                        },
                    )
                    if self.SHOULD_SHUT_DOWN:
                        return None

                    _read_smsc_interval = 62.00
                    self._log(
                        logging.DEBUG,
                        {
                            "event": "naz.Client.receive_data",
                            "stage": "end",
                            "state": "unable to read from SMSC. sleeping for {0:.2f} seconds".format(
                                _read_smsc_interval
                            ),
                            "error": str(e),
                        },
                    )
                    await asyncio.sleep(_read_smsc_interval)
                    continue  # important so that we do not hit the bug: issues/135
                chunks.append(chunk)
                bytes_recd = bytes_recd + len(chunk)

            full_pdu_data = command_length_header_data + b"".join(chunks)
            self._log(
                logging.DEBUG,
                {
                    "event": "naz.Client.receive_data",
                    "stage": "end",
                    "full_pdu_data": self._msg_to_log(msg=full_pdu_data),
                },
            )
            await self._parse_response_pdu(full_pdu_data)
            self._log(logging.INFO, {"event": "naz.Client.receive_data", "stage": "end"})
            if TESTING:
                # offer escape hatch for tests to come out of endless loop
                return full_pdu_data

    async def _parse_response_pdu(self, pdu: bytes) -> None:
        """
        Take the bytes that have been read from network and parse them into their corresponding PDU.
        The resulting PDU is then handed over to :func:`command_handlers <Client.command_handlers>`

        Parameters:
            pdu: PDU in bytes, that have been read from network
        """
        log_pdu = self._msg_to_log(msg=pdu)
        self._log(
            logging.DEBUG,
            {"event": "naz.Client._parse_response_pdu", "stage": "start", "pdu": log_pdu},
        )

        header_data = pdu[:16]
        body_data = pdu[16:]
        command_id_header_data = header_data[4:8]
        command_status_header_data = header_data[8:12]
        sequence_number_header_data = header_data[12:16]

        try:
            command_id = struct.unpack(">I", command_id_header_data)[0]
            command_status = struct.unpack(">I", command_status_header_data)[0]
            sequence_number = struct.unpack(">I", sequence_number_header_data)[0]
        except (struct.error, IndexError) as e:
            # see: https://github.com/komuw/naz/issues/135
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client._parse_response_pdu",
                    "stage": "end",
                    "state": "parse SMSC response error.",
                    "error": str(e),
                    "pdu": log_pdu,
                },
            )
            # close connection
            await self._unbind_and_disconnect()
            return None

        smpp_command = self._search_by_command_id_code(command_id)
        if not smpp_command:
            err = ValueError("command_id:{0} is unknown.".format(command_id))
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client._parse_response_pdu",
                    "stage": "end",
                    "state": "command_id:{0} is unknown.".format(command_id),
                    "error": str(err),
                },
            )
            return None

        # get associated user supplied log_id if any
        try:
            log_id, hook_metadata = await self.correlation_handler.get(
                smpp_command=smpp_command, sequence_number=sequence_number
            )
        except Exception as e:
            log_id, hook_metadata = "", ""
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client._parse_response_pdu",
                    "stage": "start",
                    "log_id": log_id,
                    "state": "correlater get error",
                    "error": str(e),
                },
            )

        await self.command_handlers(
            body_data=body_data,
            smpp_command=smpp_command,
            command_status_value=command_status,
            sequence_number=sequence_number,
            log_id=log_id,
            hook_metadata=hook_metadata,
        )
        self._log(
            logging.DEBUG,
            {
                "event": "naz.Client._parse_response_pdu",
                "stage": "end",
                "smpp_command": smpp_command,
                "log_id": log_id,
                "command_status": command_status,
            },
        )

    async def command_handlers(
        self,
        body_data: bytes,
        smpp_command: str,
        command_status_value: int,
        sequence_number: int,
        log_id: str,
        hook_metadata: str,
    ) -> None:
        """
        This routes the various different SMPP PDU to their respective handlers.

        Parameters:
            body_data: PDU body
            smpp_command: type of PDU been sent. eg bind_transceiver
            command_status_value: the response code from SMSC for a specific PDU
            sequence_number: SMPP sequence_number
            log_id: a unique identify of this request
            hook_metadata: additional metadata that you would like to be passed on to hooks
        """
        commandStatus = self._search_by_command_status_value(
            command_status_value=command_status_value
        )
        if not commandStatus:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.command_handlers",
                    "stage": "start",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "error": "command_status: `{0}` is unknown.".format(command_status_value),
                },
            )
            return None
        elif commandStatus.value != SmppCommandStatus.ESME_ROK.value:
            # we got an error from SMSC
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.command_handlers",
                    "stage": "start",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "command_status": commandStatus.value,
                    "state": commandStatus.description,
                },
            )
        else:
            self._log(
                logging.INFO,
                {
                    "event": "naz.Client.command_handlers",
                    "stage": "start",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "command_status": commandStatus.value,
                    "state": commandStatus.description,
                },
            )

        try:
            # call throttling handler
            if commandStatus.value == SmppCommandStatus.ESME_ROK.value:
                await self.throttle_handler.not_throttled()
            elif commandStatus.value in [
                SmppCommandStatus.ESME_RTHROTTLED.value,
                SmppCommandStatus.ESME_RMSGQFUL.value,
            ]:
                await self.throttle_handler.throttled()
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.command_handlers",
                    "stage": "end",
                    "error": str(e),
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": commandStatus.description,
                },
            )

        if smpp_command in [
            SmppCommand.BIND_TRANSCEIVER,
            SmppCommand.UNBIND_RESP,
            SmppCommand.SUBMIT_SM,  # We dont expect SMSC to send `submit_sm` to us.
            SmppCommand.DELIVER_SM_RESP,
            # we will never send a deliver_sm request to SMSC, which means we never
            # have to handle deliver_sm_resp
            SmppCommand.ENQUIRE_LINK_RESP,
            SmppCommand.GENERIC_NACK,  # we can ignore this
        ]:
            # we never have to handle this
            pass
        elif smpp_command == SmppCommand.BIND_TRANSCEIVER_RESP:
            # the body of `bind_transceiver_resp` only has `system_id` which is a
            # C-Octet String of variable length upto 16 octets
            if commandStatus.value == SmppCommandStatus.ESME_ROK.value:
                self.current_session_state = SmppSessionState.BOUND_TRX
        elif smpp_command == SmppCommand.UNBIND:
            # we need to handle this since we need to send unbind_resp
            # it has no body
            await self.unbind_resp(sequence_number=sequence_number)
        elif smpp_command == SmppCommand.SUBMIT_SM_RESP:
            # the body of this only has `message_id` which is a C-Octet String of variable length upto 65 octets.
            # This field contains the SMSC message_id of the submitted message.
            # It may be used at a later stage to query the status of a message, cancel
            # or replace the message.
            _message_id = body_data.replace(chr(0).encode(), b"")
            smsc_message_id = self.codec_class.decode(
                _message_id, self.encoding, self.codec_errors_level
            )
            try:
                await self.correlation_handler.put(
                    smpp_command=smpp_command,
                    sequence_number=sequence_number,
                    smsc_message_id=smsc_message_id,
                    log_id=log_id,
                    hook_metadata=hook_metadata,
                )
            except Exception as e:
                self._log(
                    logging.ERROR,
                    {
                        "event": "naz.Client.command_handlers",
                        "stage": "end",
                        "smpp_command": smpp_command,
                        "log_id": log_id,
                        "state": "correlater put error",
                        "error": str(e),
                    },
                )
        elif smpp_command == SmppCommand.DELIVER_SM:
            # HEADER::
            # command_length, int, 4octet
            # command_id, int, 4octet. `deliver_sm`
            # command_status, int, 4octet. Unused, Set to NULL.
            # sequence_number, int, 4octet. The associated `deliver_sm_resp` PDU should echo the same sequence_number.

            # BODY::
            # see section 4.6.1 of smpp v3.4 spec
            # we want to handle this pdu, bcoz we are expected to send back deliver_sm_resp
            # the body of this has the following params
            # service_type, C-Octet String, max 6 octets
            # source_addr_ton, Int, 1 octet, can be NULL
            # source_addr_npi, Int, 1 octet, can be NULL
            # source_addr, C-Octet String, max 21 octet, can be NULL
            # dest_addr_ton, Int, 1 octet
            # dest_addr_npi, Int, 1 octet
            # destination_addr,  C-Octet String, max 21 octet
            # esm_class, Int, 1 octet
            # protocol_id, Int, 1 octet
            # priority_flag, Int, 1 octet
            # schedule_delivery_time, C-Octet String, 1 octet, must be set to NULL.
            # validity_period, C-Octet String, 1 octet, must be set to NULL.
            # registered_delivery, Int, 1 octet
            # replace_if_present_flag, Int, 1 octet, must be set to NULL.
            # data_coding, Int, 1 octet
            # sm_default_msg_id, Int, 1 octet, must be set to NULL.
            # sm_length, Int, 1 octet.It is length of short message user data in octets.
            # short_message, C-Octet String, 0-254 octet

            # NB: user's hook has already been called.
            await self.deliver_sm_resp(sequence_number=sequence_number)
            try:
                # get associated user supplied log_id if any
                target_tag = struct.pack(">H", SmppOptionalTag.receipted_message_id)
                if target_tag in body_data:
                    # the PDU contains a `receipted_message_id` TLV optional tag
                    position_of_target_tag = body_data.find(target_tag)
                    # since a tag is 2 integer in size lets skip one more.
                    end_of_target_tag = position_of_target_tag + 1
                    # since after a tag, comes a tag_length which is 2 integer in size
                    # lets also skip that
                    end_of_target_tag_length = end_of_target_tag + 2
                    end_of_target_tag_length = (
                        end_of_target_tag_length + 1
                    )  # because of c-octet string null termination
                    # tag_value is of size 1 - 65
                    end_of_tag_value = end_of_target_tag_length + 65
                    tag_value = body_data[end_of_target_tag_length:end_of_tag_value]
                    _tag_value = tag_value.replace(
                        chr(0).encode(), b""
                    )  # change variable names to make mypy happy
                    t_value = self.codec_class.decode(
                        _tag_value, self.encoding, self.codec_errors_level
                    )
                    log_id, hook_metadata = await self.correlation_handler.get(
                        smpp_command=smpp_command,
                        sequence_number=sequence_number,
                        smsc_message_id=t_value,
                    )
            except Exception as e:
                log_id, hook_metadata = "", ""
                self._log(
                    logging.ERROR,
                    {
                        "event": "naz.Client.command_handlers",
                        "stage": "start",
                        "log_id": log_id,
                        "state": "correlater get error",
                        "error": str(e),
                    },
                )
        elif smpp_command == SmppCommand.ENQUIRE_LINK:
            # we have to handle this. we have to return enquire_link_resp
            # it has no body
            await self.enquire_link_resp(sequence_number=sequence_number)
        else:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.command_handlers",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "command_status": commandStatus.code,
                    "state": commandStatus.description,
                    "error": "the smpp_command: `{0}` has not been implemented in naz. please create a github issue".format(
                        smpp_command
                    ),
                },
            )

        # call user's hook for responses
        try:
            await self.hook.response(
                smpp_command=smpp_command,
                log_id=log_id,
                hook_metadata=hook_metadata,
                smsc_response=commandStatus,
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.command_handlers",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "response hook error",
                    "error": str(e),
                },
            )

    async def unbind(self) -> None:
        """
        send an UNBIND pdu to SMSC.
        """
        smpp_command = SmppCommand.UNBIND
        log_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=17))
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.unbind",
                "stage": "start",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )
        # body
        body = b""

        # header
        command_length = 16 + len(body)  # 16 is for headers
        command_id = self.command_ids[smpp_command]
        command_status = 0x00000000  # not used for `unbind`
        try:
            sequence_number = self.sequence_generator.next_sequence()
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.unbind",
                    "stage": "end",
                    "error": str(e),
                    "log_id": log_id,
                    "smpp_command": smpp_command,
                },
            )
        if sequence_number > self.max_sequence_number:
            # prevent third party sequence_generators from ruining our party
            raise ValueError(
                "the sequence_number: {0} is greater than the max: {1} allowed by SMPP spec.".format(
                    sequence_number, self.max_sequence_number
                )
            )

        try:
            await self.correlation_handler.put(
                smpp_command=smpp_command,
                sequence_number=sequence_number,
                log_id=log_id,
                hook_metadata="",
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client.unbind",
                    "stage": "end",
                    "smpp_command": smpp_command,
                    "log_id": log_id,
                    "state": "correlater put error",
                    "error": str(e),
                },
            )

        header = struct.pack(">IIII", command_length, command_id, command_status, sequence_number)
        full_pdu = header + body
        # dont queue unbind in SimpleOutboundQueue since we dont want it to be behind 10k msgs etc
        await self.send_data(smpp_command=smpp_command, msg=full_pdu, log_id=log_id)
        self._log(
            logging.INFO,
            {
                "event": "naz.Client.unbind",
                "stage": "end",
                "log_id": log_id,
                "smpp_command": smpp_command,
            },
        )

    async def shutdown(self) -> None:
        """
        Cleanly shutdown this client.
        """
        self._log(
            logging.INFO,
            {"event": "naz.Client.shutdown", "stage": "start", "state": "intiating shutdown"},
        )
        self.SHOULD_SHUT_DOWN = True

        await self._unbind_and_disconnect()

        # sleep so that client can:
        # - stop consuming from queue
        # - finish sending any SMSes it may have already picked from queue
        # - stop sending `enquire_link` requests
        # - send unbind to SMSC
        await asyncio.sleep(self.drain_duration)  # asyncio.sleep so that we do not block eventloop
        self._log(logging.DEBUG, {"event": "naz.Client.shutdown", "stage": "end"})

    async def _unbind_and_disconnect(self):
        """
        unbind from SMSC and close network connection.
        This is usually done in two situations;
          - when shutting down a naz client
          - if we got into an unrecoverable state and need to start over; issues/135
        """
        self._log(logging.DEBUG, {"event": "naz.Client._unbind_and_disconnect", "stage": "start"})

        if typing.TYPE_CHECKING:
            # make mypy happy; https://github.com/python/mypy/issues/4805
            assert isinstance(self.writer, asyncio.streams.StreamWriter)
            assert isinstance(self.writer.transport, asyncio.transports.Transport)
        try:
            # 1. set buffers to 0
            # 2. unbind
            # 3. drain
            # 4. close connection
            # in that order

            # see: https://github.com/komuw/naz/issues/117
            self.writer.transport.set_write_buffer_limits(0)  # pytype: disable=attribute-error
            # https://github.com/google/pytype/issues/350
            await self.unbind()
            async with self.drain_lock:
                await self.writer.drain()
            self.writer.close()
        except (
            ConnectionError,
            TimeoutError,
            asyncio.TimeoutError,
            socket.error,
            socket.herror,
            socket.gaierror,
            socket.timeout,
        ) as e:
            self._log(
                logging.ERROR,
                {
                    "event": "naz.Client._unbind_and_disconnect",
                    "stage": "end",
                    "state": "unable to write to SMSC",
                    "error": str(e),
                },
            )
        self._log(logging.DEBUG, {"event": "naz.Client._unbind_and_disconnect", "stage": "end"})


class NazClientError(Exception):
    """
    Error raised when there's an error instantiating a naz Client.
    """

    pass



import signal
import asyncio
import logging
import functools

import naz


async def _signal_handling(
    logger: naz.log.BaseLogger, client: naz.Client, loop: asyncio.events.AbstractEventLoop
) -> None:
    try:
        for _signal in [signal.SIGHUP, signal.SIGQUIT, signal.SIGTERM]:
            loop.add_signal_handler(
                _signal,
                functools.partial(
                    asyncio.ensure_future,
                    _handle_termination_signal(logger=logger, _signal=_signal, client=client),
                ),
            )
    except ValueError as e:
        logger.log(
            logging.DEBUG,
            {
                "event": "naz.cli.signals",
                "stage": "end",
                "state": "this OS does not support the said signal",
                "error": str(e),
            },
        )


async def _handle_termination_signal(
    logger: naz.log.BaseLogger, _signal: "signal.Signals", client: naz.Client
) -> None:
    logger.log(
        logging.INFO,
        {
            "event": "naz.cli.signals",
            "stage": "start",
            "state": "received termination signal",
            "signal": _signal.name,
        },
    )

    await client.shutdown()

    logger.log(
        logging.INFO,
        {"event": "naz.cli.signals", "stage": "end", "state": "client has succesfully shutdown"},
    )
    return



import abc
import time
import typing
import asyncio
import logging

from . import log


class BaseRateLimiter(abc.ABC):
    """
    This is the interface that must be implemented to satisfy naz's rate limiting.
    User implementations should inherit this class and
    implement the :func:`limit <BaseRateLimiter.limit>` methods with the type signatures shown.

    It may be important to control the rate at which the client(naz) sends requests to an SMSC/server.
    naz lets you do this, by allowing you to specify a custom rate limiter.
    """

    @abc.abstractmethod
    async def limit(self) -> None:
        """
        rate limit sending of messages to SMSC.
        """
        raise NotImplementedError("limit method must be implemented.")


class SimpleRateLimiter(BaseRateLimiter):
    """
    This is an implementation of BaseRateLimiter.

    It does rate limiting using a `token bucket rate limiting algorithm <https://en.wikipedia.org/wiki/Token_bucket>`_

    example usage:

    .. highlight:: python
    .. code-block:: python

        rateLimiter = SimpleRateLimiter(send_rate=10)
        await rateLimiter.limit()
        send_messsages()
    """

    def __init__(
        self, send_rate: float = 100_000.00, logger: typing.Union[None, log.BaseLogger] = None
    ) -> None:
        """
        Parameters:
            send_rate: the maximum rate, in messages/second, at which naz can send messages to SMSC.
        """
        if not isinstance(send_rate, float):
            raise ValueError(
                "`send_rate` should be of type:: `float` You entered: {0}".format(type(send_rate))
            )
        if not isinstance(logger, (type(None), log.BaseLogger)):
            raise ValueError(
                "`logger` should be of type:: `None` or `naz.log.BaseLogger` You entered: {0}".format(
                    type(logger)
                )
            )

        self.send_rate: float = send_rate
        self.max_tokens: float = self.send_rate
        self.tokens: float = self.max_tokens
        self.delay_for_tokens: float = 1.0
        self.updated_at: float = time.monotonic()

        self.messages_delivered: int = 0
        self.effective_send_rate: float = 0.00
        if logger is not None:
            self.logger = logger
        else:
            self.logger = log.SimpleLogger("naz.SimpleRateLimiter")

    async def limit(self) -> None:
        self.logger.log(logging.DEBUG, {"event": "naz.SimpleRateLimiter.limit", "stage": "start"})
        while self.tokens < 1:
            self._add_new_tokens()
            # todo: sleep in an exponetial manner upto a maximum then wrap around.
            await asyncio.sleep(self.delay_for_tokens)
            self.logger.log(
                logging.INFO,
                {
                    "event": "naz.SimpleRateLimiter.limit",
                    "stage": "end",
                    "state": "limiting rate",
                    "send_rate": self.send_rate,
                    "delay": self.delay_for_tokens,
                    "effective_send_rate": self.effective_send_rate,
                },
            )

        self.messages_delivered += 1
        self.tokens -= 1

    def _add_new_tokens(self) -> None:
        now = time.monotonic()
        time_since_update = now - self.updated_at
        self.effective_send_rate = self.messages_delivered / time_since_update
        new_tokens = time_since_update * self.send_rate
        if new_tokens > 1:
            self.tokens = min(self.tokens + new_tokens, self.max_tokens)
            self.updated_at = now
            self.messages_delivered = 0



# The code in this file is copied from https://github.com/praekelt/vumi/blob/master/vumi/codecs/vumi_codecs.py
# which is in turn largely copied from http://stackoverflow.com/questions/13130935/decode-7-bit-gsm
# Vumi's license is included below:

# Copyright (c) Praekelt Foundation and individual contributors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1.  Redistributions of source code must retain the above copyright notice,
#         this list of conditions and the following disclaimer.

#     2.  Redistributions in binary form must reproduce the above copyright
#         notice, this list of conditions and the following disclaimer in the
#         documentation and/or other materials provided with the distribution.

#     3.  Neither the name of the Praekelt Foundation nor the names of its
#         contributors may be used to endorse or promote products derived from
#         this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import abc
import sys
import codecs


# An alternative to using this codec module is to use: https://github.com/dsch/gsm0338
# however, I'm guessing that vumi has been in use longer and we should thus go with it.


class NazCodecException(Exception):
    pass


class GSM7BitCodec(codecs.Codec):
    """
    """

    gsm_basic_charset = (
        "@£$¥èéùìòÇ\nØø\rÅåΔ_ΦΓΛΩΠΨΣΘΞ\x1bÆæßÉ !\"#¤%&'()*+,-./0123456789:;"
        "<=>?¡ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÑÜ`¿abcdefghijklmnopqrstuvwxyzäö"
        "ñüà"
    )

    gsm_basic_charset_map = dict((l, i) for i, l in enumerate(gsm_basic_charset))

    gsm_extension = (
        "````````````````````^```````````````````{}`````\\````````````[~]`"
        "|````````````````````````````````````€``````````````````````````"
    )

    gsm_extension_map = dict((l, i) for i, l in enumerate(gsm_extension))

    def encode(self, input, errors="strict"):
        """
        errors can be 'strict', 'replace' or 'ignore'
        eg:
            xcodec.encode("Zoë","gsm0338") will fail with UnicodeEncodeError
            but
            xcodec.encode("Zoë","gsm0338", 'replace') will return b'Zo?'
            and
            xcodec.encode("Zoë","gsm0338", 'ignore') will return b'Zo'
        """
        result = []
        for position, c in enumerate(input):
            idx = self.gsm_basic_charset_map.get(c)
            if idx is not None:
                result.append(chr(idx))
                continue
            idx = self.gsm_extension_map.get(c)
            if idx is not None:
                result.append(chr(27) + chr(idx))
            else:
                result.append(self.handle_encode_error(c, errors, position, input))

        obj = "".join(result)
        # this is equivalent to;
        # import six; six.b('someString')
        # see:
        # https://github.com/benjaminp/six/blob/68112f3193c7d4bef5ad86ed1b6ed528edd9093d/six.py#L625
        obj = obj.encode("latin-1")
        return (obj, len(obj))

    def handle_encode_error(self, char, handler_type, position, obj):
        handler = getattr(self, "handle_encode_%s_error" % (handler_type,), None)
        if handler is None:
            raise NazCodecException("Invalid errors type %s for GSM7BitCodec", handler_type)
        return handler(char, position, obj)

    @staticmethod
    def handle_encode_strict_error(char, position, obj):
        raise UnicodeEncodeError("gsm0338", char, position, position + 1, repr(obj))

    @staticmethod
    def handle_encode_ignore_error(char, position, obj):
        return ""

    def handle_encode_replace_error(self, char, position, obj):
        return chr(self.gsm_basic_charset_map.get("?"))

    def decode(self, input, errors="strict"):
        """
        errors can be 'strict', 'replace' or 'ignore'
        """
        res = iter(input)
        result = []
        for position, c in enumerate(res):
            try:
                if c == 27:
                    c = next(res)
                    result.append(self.gsm_extension[c])
                else:
                    result.append(self.gsm_basic_charset[c])
            except IndexError as indexErrorException:
                result.append(
                    self.handle_decode_error(c, errors, position, input, indexErrorException)
                )

        obj = "".join(result)
        return (obj, len(obj))

    def handle_decode_error(self, char, handler_type, position, obj, indexErrorException):
        handler = getattr(self, "handle_decode_%s_error" % (handler_type,), None)
        if handler is None:
            raise NazCodecException("Invalid errors type %s for GSM7BitCodec", handler_type)
        return handler(char, position, obj, indexErrorException)

    @staticmethod
    def handle_decode_strict_error(char, position, obj, indexErrorException):
        # https://github.com/google/pytype/issues/349
        raise UnicodeDecodeError(
            "gsm0338",
            chr(char).encode("latin-1"),
            position,
            position + 1,
            repr(obj),  # pytype: disable=wrong-arg-types
        ) from indexErrorException

    @staticmethod
    def handle_decode_ignore_error(char, position, obj, indexErrorException):
        return ""

    @staticmethod
    def handle_decode_replace_error(char, position, obj, indexErrorException):
        return "?"


class UCS2Codec(codecs.Codec):
    """
    UCS2 is for all intents & purposes assumed to be the same as
    big endian UTF16.
    """

    def encode(self, input, errors="strict"):
        # https://github.com/google/pytype/issues/348
        return codecs.utf_16_be_encode(input, errors)  # pytype: disable=module-attr

    def decode(self, input, errors="strict"):
        return codecs.utf_16_be_decode(input, errors)  # pytype: disable=module-attr


class BaseNazCodec(abc.ABC):
    """
    This is the interface that must be implemented to satisfy naz's encoding/decoding.
    User implementations should inherit this class and
    implement the :func:`encode <BaseNazCodec.encode>` and :func:`decode <BaseNazCodec.decode>` methods with the type signatures shown.

    naz calls an implementation of this class to encode/decode messages.
    """

    @abc.abstractmethod
    def encode(self, string_to_encode: str, encoding: str, errors: str) -> bytes:
        """
        return an encoded version of the string as a bytes object

        Parameters:
            string_to_encode: the string to encode
            encoding: encoding scheme. eg utf-8, gsm0338 etc
            errors: `same as defined in pythons codec.encode <https://docs.python.org/3/library/codecs.html#codecs.encode>`_

        Returns:
            encoded version of input string
        """
        raise NotImplementedError("encode method must be implemented.")

    @abc.abstractmethod
    def decode(self, byte_string: bytes, encoding: str, errors: str) -> str:
        """
        return a string decoded from the given bytes.

       Parameters:
            byte_string: the bytes to decode
            encoding: encoding scheme. eg utf-8, gsm0338 etc
            errors: `same as defined in pythons codec.decode <https://docs.python.org/3/library/codecs.html#codecs.decode>`_
        """
        raise NotImplementedError("decode method must be implemented.")


class SimpleNazCodec(BaseNazCodec):
    """
    This is an implementation of BaseNazCodec.

    SMPP uses a 7-bit GSM character set. This class implements that encoding/decoding scheme.
    This class can also be used with the usual `python standard encodings <https://docs.python.org/3/library/codecs.html#standard-encodings>`_

    example usage:

    .. highlight:: python
    .. code-block:: python

       ncodec = SimpleNazCodec()

       ncodec.encode("Zoë", "utf-16be")
       ncodec.encode("Zoë", "utf-8")
       ncodec.encode("HÜLK", 'gsm0338')

       ncodec.decode(b'Zo\xc3\xab', 'gsm0338', 'ignore')
       ncodec.decode(b'Zo\xc3\xab', 'utf8')
    """

    custom_codecs = {"gsm0338": GSM7BitCodec(), "ucs2": UCS2Codec()}

    def encode(self, string_to_encode: str, encoding: str, errors: str) -> bytes:
        if not isinstance(string_to_encode, str):
            raise NazCodecException("Only strings accepted for encoding.")
        encoding = encoding or sys.getdefaultencoding()
        if encoding in self.custom_codecs:
            encoder = self.custom_codecs[encoding].encode
        else:
            encoder = codecs.getencoder(encoding)
        obj, _ = encoder(string_to_encode, errors)
        return obj

    def decode(self, byte_string: bytes, encoding: str, errors: str) -> str:
        if not isinstance(byte_string, (bytes, bytearray)):
            raise NazCodecException("Only bytestrings accepted for decoding.")
        encoding = encoding or sys.getdefaultencoding()
        if encoding in self.custom_codecs:
            decoder = self.custom_codecs[encoding].decode
        else:
            decoder = codecs.getdecoder(encoding)
        obj, _ = decoder(byte_string, errors)
        return obj



import abc
import asyncio
import typing


class BaseOutboundQueue(abc.ABC):
    """
    This is the interface that must be implemented to satisfy naz's outbound queue.
    User implementations should inherit this class and
    implement the :func:`enqueue <BaseOutboundQueue.enqueue>` and :func:`dequeue <BaseOutboundQueue.dequeue>` methods with the type signatures shown.

    naz calls an implementation of this class to enqueue and/or dequeue an item.
    """

    @abc.abstractmethod
    async def enqueue(self, item: dict) -> None:
        """
        enqueue/save an item.

        Parameters:
            item: The item to be enqueued/saved
        """
        raise NotImplementedError("enqueue method must be implemented.")

    @abc.abstractmethod
    async def dequeue(self) -> typing.Dict[typing.Any, typing.Any]:
        """
        dequeue an item.

        Returns:
            item that was dequeued
        """
        raise NotImplementedError("dequeue method must be implemented.")


class SimpleOutboundQueue(BaseOutboundQueue):
    """
    This is an in-memory implementation of BaseOutboundQueue.

    Note: It should only be used for tests and demo purposes.
    """

    def __init__(self, maxsize: int = 2500) -> None:
        """
        Parameters:
            maxsize: the maximum number of items(not size) that can be put in the queue.
        """
        if not isinstance(maxsize, int):
            raise ValueError(
                "`maxsize` should be of type:: `int` You entered: {0}".format(type(maxsize))
            )
        self.queue: asyncio.queues.Queue = asyncio.Queue(maxsize=maxsize)

    async def enqueue(self, item: dict) -> None:
        self.queue.put_nowait(item)

    async def dequeue(self) -> typing.Dict[typing.Any, typing.Any]:
        return await self.queue.get()



import abc
import time
import typing


class BaseCorrelater(abc.ABC):
    """
    Interface that must be implemented to satisfy naz's Correlater.
    User implementations should inherit this class and
    implement the :func:`get <BaseCorrelater.get>` and :func:`put <BaseCorrelater.put>` methods with the type signatures shown.

    A Correlater is class that naz uses to store relations between SMPP sequence numbers
    and user applications' log_id's and/or hook_metadata.

    | Note: This correlation is on a BEST effort basis; it is not guaranteed to be reliable.
    | One reason, among others, is that the SMPP specifiation mandates sequence numbers to wrap around after ≈ 2billion.
    | Another reason is that we use receipted_message_id tag from deliver_sm to correlate a delivery notification
      with a submit_sm_resp. However receipted_message_id is an optional parameter that SMSC may omit.
    """

    @abc.abstractmethod
    async def put(
        self,
        smpp_command: str,
        sequence_number: int,
        log_id: str,
        hook_metadata: str,
        smsc_message_id: typing.Union[None, str] = None,
    ) -> None:
        """
        called by naz to put/store the correlation of a given SMPP sequence number to log_id and/or hook_metadata.

        Parameters:
            smpp_command: any one of the SMSC commands eg submit_sm
            sequence_number: SMPP sequence_number
            log_id: an ID that a user's application had previously supplied to naz to track/correlate different messages.
            hook_metadata: a string that a user's application had previously supplied to naz that it may want to be correlated with the log_id.
            smsc_message_id: a unique identifier of a particular message on the SMSC. It comes from SMSC
        """
        raise NotImplementedError("put method must be implemented.")

    @abc.abstractmethod
    async def get(
        self,
        smpp_command: str,
        sequence_number: int,
        smsc_message_id: typing.Union[None, str] = None,
    ) -> typing.Tuple[str, str]:
        """
        called by naz to get the correlation of a given SMPP sequence number to log_id and/or hook_metadata.

        Parameters:
            smpp_command: any one of the SMSC commands eg submit_sm
            sequence_number: SMPP sequence_number
            smsc_message_id: a unique identifier of a particular message on the SMSC. It comes from SMSC

        Returns:
            log_id and hook_metadata
        """
        raise NotImplementedError("get method must be implemented.")


class SimpleCorrelater(BaseCorrelater):
    """
    A simple implementation of BaseCorrelater.
    It stores the correlation/relation between a given SMPP sequence_number(and/or smsc_message_id) and a user supplied log_id and/or hook_metadata.

    SimpleCorrelater also features an auto-expiration of dictionary keys(and their values) based on time.

     The storage is done in memory using a python dictionary. The storage looks like:

    .. highlight:: python
    .. code-block:: python

       {
            "sequence_number1": {
                "log_id": "log_id1",
                "hook_metadata": "hook_metadata1",
                "stored_at": 681.109023565
            },
            "smsc_message_id1": {
                "log_id": "log_id1",
                "hook_metadata": "hook_metadata1",
                "stored_at": 681.109023565
            },
           "sequence_number1": {
            "log_id": "log_id2",
            "hook_metadata": "hook_metadata2",
            "stored_at": 682.109023565
           }
           ...
        }
    """

    def __init__(self, max_ttl: float = 15.00) -> None:
        """
        Parameters:
            max_ttl: The time in seconds that an item is going to be stored.
                    After the expiration of max_ttl seconds, that item will be deleted.
        """
        if not isinstance(max_ttl, float):
            raise ValueError(
                "`max_ttl` should be of type:: `float` You entered: {0}".format(type(max_ttl))
            )
        self.store: dict = {}
        self.max_ttl: float = max_ttl

    async def put(
        self,
        smpp_command: str,
        sequence_number: int,
        log_id: str,
        hook_metadata: str,
        smsc_message_id: typing.Union[None, str] = None,
    ) -> None:
        stored_at = time.monotonic()
        if smpp_command == "submit_sm_resp":
            # TODO: dict with smsc_message_id should replace dict with corresponding sequence_number
            # currently we are not deduping data; we should
            self.store[smsc_message_id] = {
                "log_id": log_id,
                "hook_metadata": hook_metadata,
                "stored_at": stored_at,
            }
        else:
            self.store[sequence_number] = {
                "log_id": log_id,
                "hook_metadata": hook_metadata,
                "stored_at": stored_at,
            }

        # garbage collect
        await self.delete_after_ttl()

    async def get(
        self,
        smpp_command: str,
        sequence_number: int,
        smsc_message_id: typing.Union[None, str] = None,
    ) -> typing.Tuple[str, str]:
        if smpp_command == "deliver_sm":
            item = self.store.get(smsc_message_id)
        else:
            item = self.store.get(sequence_number)
        if not item:
            # garbage collect
            await self.delete_after_ttl()
            return "", ""

        # garbage collect
        await self.delete_after_ttl()
        return item["log_id"], item["hook_metadata"]

    async def delete_after_ttl(self) -> None:
        """
        iterate over all stored items and delete any that are
        older than self.max_ttl seconds
        """
        now = time.monotonic()
        for key in list(self.store.keys()):
            stored_at = self.store[key]["stored_at"]
            time_diff = now - stored_at
            if time_diff > self.max_ttl:
                del self.store[key]



import os
import sys
import logging
import importlib

import naz


def load_class(dotted_path: str, logger: naz.log.BaseLogger) -> naz.Client:
    """
    taken from: https://github.com/coleifer/huey/blob/4138d454cc6fd4d252c9350dbd88d74dd3c67dcb/huey/utils.py#L44
    huey is released under MIT license a copy of which can be found at: https://github.com/coleifer/huey/blob/master/LICENSE

    The license is also included below:

    Copyright (c) 2017 Charles Leifer

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
    """
    try:
        path, klass = dotted_path.rsplit(".", 1)
        mod = importlib.import_module(path)
        attttr = getattr(mod, klass)
        return attttr
    except Exception as e:
        cur_dir = os.getcwd()
        if cur_dir not in sys.path:
            sys.path.insert(0, cur_dir)
            return load_class(dotted_path, logger)
        err_message = "Error importing {0}".format(dotted_path)
        logger.log(
            logging.ERROR,
            {"event": "naz.cli.main", "stage": "end", "state": err_message, "error": str(e)},
        )
        raise e



from .client import Client  # noqa: F401

from . import q  # noqa: F401
from . import throttle  # noqa: F401
from . import sequence  # noqa: F401
from . import log  # noqa: F401
from . import nazcodec  # noqa: F401
from . import correlater  # noqa: F401
from . import ratelimiter  # noqa: F401

from .state import (  # noqa: F401
    SmppSessionState,
    SmppCommand,
    CommandStatus,
    SmppCommandStatus,
    DataCoding,
    SmppDataCoding,
    SmppOptionalTag,
)

from . import __version__  # noqa: F401



import abc
import time
import typing
import logging
import collections


class BaseLogger(abc.ABC):
    """
    Interface that must be implemented to satisfy naz's logger.
    User implementations should inherit this class and
    implement the :func:`bind <BaseLogger.bind>` and :func:`log <BaseLogger.log>` methods with the type signatures shown.

    A logger is class with methods that are called whenever naz wants to log something.
    This enables developers to implement logging in any way that they want.
    """

    @abc.abstractmethod
    def bind(self, level: typing.Union[str, int], log_metadata: dict) -> None:
        """
        called when a naz client is been instantiated so that the logger can be
        notified of loglevel & log_metadata that a user supplied to a naz client.
        The logger can choose to bind these log_metadata to itself.

        Parameters:
            loglevel: logging level eg DEBUG
            log_metadata: log metadata that can be included in all log statements
        """
        raise NotImplementedError("`bind` method must be implemented.")

    @abc.abstractmethod
    def log(self, level: typing.Union[str, int], log_data: dict) -> None:
        """
        called by naz everytime it wants to log something.

        Parameters:
            level: logging level eg `logging.INFO`
            log_data: the message to log
        """
        raise NotImplementedError("`log` method must be implemented.")


class SimpleLogger(BaseLogger):
    """
    This is an implementation of BaseLogger.
    It implements a structured logger that renders logs as a dict.

    example usage:

    .. highlight:: python
    .. code-block:: python

        logger = SimpleLogger("myLogger")
        logger.bind(level="INFO",
                    log_metadata={"customer_id": "34541"})
        logger.log(logging.INFO,
                   {"event": "web_request", "url": "https://www.google.com/"})
    """

    def __init__(self, logger_name: str, handler: logging.Handler = logging.StreamHandler()):
        """
        Parameters:
            logger_name: name of the logger. it should be unique per logger.
        """
        if not isinstance(logger_name, str):
            raise ValueError(
                "`logger_name` should be of type:: `str` You entered: {0}".format(type(logger_name))
            )
        if not isinstance(handler, logging.Handler):
            raise ValueError(
                "`handler` should be of type:: `logging.Handler` You entered: {0}".format(
                    type(handler)
                )
            )

        self.logger_name = logger_name
        self.handler = handler
        self.logger: typing.Union[None, logging.LoggerAdapter] = None

    def bind(self, level: typing.Union[str, int], log_metadata: dict) -> None:
        level = self._nameToLevel(level=level)

        self._logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter("%(message)s")
        self.handler.setFormatter(formatter)
        self.handler.setLevel(level)
        if not self._logger.handlers:
            self._logger.addHandler(self.handler)
        self._logger.setLevel(level)
        self.logger = _NazLoggingAdapter(self._logger, log_metadata)

    def log(self, level: typing.Union[str, int], log_data: typing.Union[str, dict]) -> None:
        level = self._nameToLevel(level=level)

        if not self.logger:
            self.bind(level=level, log_metadata={})
        if typing.TYPE_CHECKING:
            # make mypy happy; https://github.com/python/mypy/issues/4805
            assert isinstance(self.logger, logging.LoggerAdapter)

        if level >= logging.ERROR:
            self.logger.log(level, log_data, exc_info=True)
        else:
            self.logger.log(level, log_data)

    @staticmethod
    def _nameToLevel(level: typing.Union[str, int]) -> int:
        try:
            if isinstance(level, str):
                # please mypy
                _level: int = logging._nameToLevel[level.upper()]
            else:
                _level = level
            return _level
        except KeyError as e:
            raise ValueError(
                "`level` should be one of; 'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'"
            ) from e


class _NazLoggingAdapter(logging.LoggerAdapter):
    _converter = time.localtime
    _formatter = logging.Formatter()

    def process(self, msg, kwargs):
        timestamp = self.formatTime()

        if isinstance(msg, str):
            merged_msg = "{0} {1} {2}".format(timestamp, msg, self.extra)
            if self.extra == {}:
                merged_msg = "{0} {1}".format(timestamp, msg)
            return merged_msg, kwargs
        else:
            _timestamp = {"timestamp": timestamp}
            # _timestamp should appear first in resulting dict
            merged_msg = {**_timestamp, **msg, **self.extra}
            return "{0}".format(merged_msg), kwargs

    def formatTime(self):
        """
        Return the creation time of the specified log event as formatted text.

        This code is borrowed from:
        https://docs.python.org/3/library/logging.html#logging.Formatter.formatTime

        The basic behaviour is as follows: an ISO8601-like (or RFC 3339-like) format is used.
        This function uses `time.localtime()` to convert the creation time to a tuple.
        """
        now = time.time()
        msecs = (now - int(now)) * 1000

        ct = self._converter(now)
        t = time.strftime(self._formatter.default_time_format, ct)
        s = self._formatter.default_msec_format % (t, msecs)
        return s


class BreachHandler(logging.StreamHandler):
    """
    This is an implementation of `logging.Handler` that puts logs in an in-memory ring buffer.
    When a trigger condition(eg a certain log level) is met;
    then all the logs in the buffer are flushed into a given stream(file, stdout etc)

    It is a bit like
    `logging.handlers.MemoryHandler <https://docs.python.org/3.6/library/logging.handlers.html#logging.handlers.MemoryHandler>`_
    except that it does not flush when the ring-buffer capacity is met but only when/if the trigger is met.

    It is inspired by the article
    `Triggering Diagnostic Logging on Exception <https://tersesystems.com/blog/2019/07/28/triggering-diagnostic-logging-on-exception/>`_

    example usage:

    .. highlight:: python
    .. code-block:: python

        _handler = naz.log.BreachHandler()
        logger = naz.log.SimpleLogger("aha", handler=_handler)
        logger.bind(level="INFO", log_metadata={"id": "123"})

        logger.log(logging.INFO, {"name": "Jayz"})
        logger.log(logging.ERROR, {"msg": "Houston, we got 99 problems."})

        # or alternatively, to use it with python stdlib logger
        logger = logging.getLogger("my-logger")
        handler = naz.log.BreachHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        handler.setLevel("DEBUG")
        if not logger.handlers:
            logger.addHandler(handler)
        logger.setLevel("DEBUG")

        logger.info("I did records for Tweet before y'all could even tweet - Dr. Missy Elliot")
        logger.error("damn")
    """

    def __init__(self, trigger_level=logging.WARNING, buffer_size=10_000, stream=None):
        """
        Parameters:
            trigger_level: the log level that will trigger this handler to flush logs to :py:attr:`~stream`
            buffer_size: the maximum number of log records to store in the ring buffer
            stream: a `file like object <https://docs.python.org/3/library/io.html>`_ that can be logged to.
        """
        # call `logging.StreamHandler` init
        super(BreachHandler, self).__init__(stream=stream)
        self.trigger_level = trigger_level
        self.buffer_size = buffer_size
        self.buffered_logs = collections.deque(maxlen=self.buffer_size)
        # assuming each log record is 250 bytes, then the maximum
        # memory used by `buffered_logs` will always be == 250*10_000/(1000*1000) == 2.5MB

    def emit(self, record):
        """
        Emit a record.
        Implementation is mostly taken from `logging.StreamHandler`
        """
        try:
            # 1. append new item to deque(ring-buffer)
            msg = self.format(record)
            self.buffered_logs.append(msg)

            # 2. check if the loglevel of the current record >= `self.trigger_level`
            #    if it is.
            #      (a) acquire lock
            #      (b) stream.write and flush from `self.buffered_logs`
            #      (c) clear out `self.buffered_logs`
            if record.levelno < self.trigger_level:
                return
            else:
                stream = self.stream
                for _msg in self.buffered_logs:
                    stream.write(_msg)
                    stream.write(self.terminator)
                self.buffered_logs.clear()
                self.flush()
        except Exception:
            self.handleError(record)



import abc


class BaseSequenceGenerator(abc.ABC):
    """
    Interface that must be implemented to satisfy naz's sequence generator.
    User implementations should inherit this class and
    implement the :func:`next_sequence <BaseSequenceGenerator.next_sequence>` methods with the type signatures shown.

    In SMPP, sequence_number is an Integer which allows SMPP requests and responses to be correlated.
    The sequence_number should increase monotonically and ought to be in the range 1 - 2,147,483,647

    The sequence_number should wrap around when it reaches the maximum allowed by SMPP specification.
    """

    @abc.abstractmethod
    def next_sequence(self) -> int:
        """
        method that returns a monotonically increasing Integer in the range 1 - 2,147,483,647
        """
        raise NotImplementedError("next_sequence method must be implemented.")


class SimpleSequenceGenerator(BaseSequenceGenerator):
    """
    This is an implementation of BaseSequenceGenerator.
    """

    min_sequence_number: int = 0x00000001
    max_sequence_number: int = 0x7FFFFFFF

    def __init__(self) -> None:
        self.sequence_number: int = self.min_sequence_number

    def next_sequence(self) -> int:
        if self.sequence_number == self.max_sequence_number:
            # wrap around
            self.sequence_number = self.min_sequence_number
        else:
            self.sequence_number += 1
        return self.sequence_number



import abc
import time
import typing
import logging

from . import log


class BaseThrottleHandler(abc.ABC):
    """
    This is the interface that must be implemented to satisfy naz's throttle handling.
    User implementations should inherit this class and
    implement the :func:`throttled <BaseThrottleHandler.throttled>`, :func:`not_throttled <BaseThrottleHandler.not_throttled>`,
    :func:`allow_request <BaseThrottleHandler.allow_request>` and
    :func:`throttle_delay <BaseThrottleHandler.throttle_delay>` methods with the type signatures shown.

    When an SMPP client exceeds it's rate limit, or when the SMSC is under load or for whatever reason;
    The SMSC may decide to start throtlling requests from that particular client.
    When it does so, it replies to the client with a throttling status. Under such conditions, it is important for the client to start
    rate limiting itself. The way naz implements this self imposed self-regulation is via Throttle Handlers.

    The methods in this class are also called when the SMSC is under load and is responding with `ESME_RMSGQFUL`(message queue full) responses
    """

    @abc.abstractmethod
    async def throttled(self) -> None:
        """
        this method will be called by naz everytime we get a throttling response from SMSC.
        """
        raise NotImplementedError("throttled method must be implemented.")

    @abc.abstractmethod
    async def not_throttled(self) -> None:
        """
        this method will be called by naz everytime we get any response from SMSC that is not a throttling response.
        """
        raise NotImplementedError("not_throttled method must be implemented.")

    @abc.abstractmethod
    async def allow_request(self) -> bool:
        """
        this method will be called by naz just before sending a request to SMSC.
        The response from this method will determine wether naz will send the request or not.
        """
        raise NotImplementedError("allow_request method must be implemented.")

    @abc.abstractmethod
    async def throttle_delay(self) -> float:
        """
        if the last :func:`allow_request <BaseThrottleHandler.allow_request>` method call returned False(thus denying sending a request),
        naz will call the throttle_delay method
        to determine how long in seconds to wait before calling allow_request again.
        """
        raise NotImplementedError("throttle_delay method must be implemented.")


class SimpleThrottleHandler(BaseThrottleHandler):
    """
    This is an implementation of BaseThrottleHandler.

    It works by:

    - calculating the percentage of responses from the SMSC that are THROTTLING(or ESME_RMSGQFUL) responses.
    - if that percentage goes above :attr:`deny_request_at <SimpleThrottleHandler.deny_request_at>` percent AND \
    total number of responses from SMSC is greater than :attr:`sample_size <SimpleThrottleHandler.sample_size>` over \
    :attr:`sampling_period <SimpleThrottleHandler.sampling_period>` seconds
    - then deny making anymore requests to SMSC

    """

    def __init__(
        self,
        sampling_period: float = 180.00,
        sample_size: float = 50.00,
        deny_request_at: float = 1.00,
        throttle_wait: float = 3.00,
        logger: typing.Union[None, log.BaseLogger] = None,
    ) -> None:
        """
        Parameters:
            sampling_period: the duration in seconds over which we will calculate the percentage of throttled responses.
            sample_size: the minimum number of responses we should have got from SMSC over :sampling_period duration to enable us make a decision.
            deny_request_at: the percent of throtlled responses above which we will deny naz from sending more requests to SMSC.
            throttle_wait: the time in seconds to wait before calling allow_request after the last allow_request that returned False.
        """
        if not isinstance(sampling_period, float):
            raise ValueError(
                "`sampling_period` should be of type:: `float` You entered: {0}".format(
                    type(sampling_period)
                )
            )
        if not isinstance(sample_size, float):
            raise ValueError(
                "`sample_size` should be of type:: `float` You entered: {0}".format(
                    type(sample_size)
                )
            )
        if not isinstance(deny_request_at, float):
            raise ValueError(
                "`deny_request_at` should be of type:: `float` You entered: {0}".format(
                    type(deny_request_at)
                )
            )
        if not isinstance(throttle_wait, float):
            raise ValueError(
                "`throttle_wait` should be of type:: `float` You entered: {0}".format(
                    type(throttle_wait)
                )
            )
        if not isinstance(logger, (type(None), log.BaseLogger)):
            raise ValueError(
                "`logger` should be of type:: `None` or `naz.log.BaseLogger` You entered: {0}".format(
                    type(logger)
                )
            )

        self.NON_throttle_responses: int = 0
        self.throttle_responses: int = 0
        self.updated_at: float = time.monotonic()

        self.sampling_period: float = sampling_period
        self.sample_size: float = sample_size
        self.deny_request_at: float = deny_request_at
        self.throttle_wait: float = throttle_wait

        if logger is not None:
            self.logger = logger
        else:
            self.logger = log.SimpleLogger("naz.SimpleThrottleHandler")

    @property
    def percent_throttles(self) -> float:
        total_smsc_responses: int = self.NON_throttle_responses + self.throttle_responses
        if total_smsc_responses < self.sample_size:
            # we do not have enough data to make decision, so asssume happy case
            return 0.0
        return round((self.throttle_responses / (total_smsc_responses)) * 100, 2)

    async def allow_request(self) -> bool:
        self.logger.log(
            logging.DEBUG, {"event": "naz.SimpleThrottleHandler.allow_request", "stage": "start"}
        )
        # calculat percentage of throttles before resetting NON_throttle_responses and throttle_responses
        current_percent_throttles: float = self.percent_throttles
        _throttle_responses: int = self.throttle_responses
        _NON_throttle_responses: int = self.NON_throttle_responses

        now: float = time.monotonic()
        time_since_update: float = now - self.updated_at
        if time_since_update > self.sampling_period:
            # we are only interested in percent throttles in buckets of self.sampling_period seconds.
            # so reset values after self.sampling_period seconds.
            self.NON_throttle_responses = 0
            self.throttle_responses = 0
            self.updated_at = now
        if current_percent_throttles > self.deny_request_at:
            self.logger.log(
                logging.INFO,
                {
                    "event": "naz.SimpleThrottleHandler.allow_request",
                    "stage": "end",
                    "percent_throttles": current_percent_throttles,
                    "throttle_responses": _throttle_responses,
                    "NON_throttle_responses": _NON_throttle_responses,
                    "sampling_period": self.sampling_period,
                    "sample_size": self.sample_size,
                    "deny_request_at": self.deny_request_at,
                    "state": "deny_request",
                },
            )
            return False
        self.logger.log(
            logging.DEBUG,
            {
                "event": "naz.SimpleThrottleHandler.allow_request",
                "stage": "end",
                "percent_throttles": current_percent_throttles,
                "throttle_responses": _throttle_responses,
                "NON_throttle_responses": _NON_throttle_responses,
                "sampling_period": self.sampling_period,
                "sample_size": self.sample_size,
                "deny_request_at": self.deny_request_at,
                "state": "allow_request",
            },
        )
        return True

    async def not_throttled(self) -> None:
        self.NON_throttle_responses += 1

    async def throttled(self) -> None:
        self.throttle_responses += 1

    async def throttle_delay(self) -> float:
        # todo: sleep in an exponetial manner upto a maximum then wrap around.
        return self.throttle_wait



import typing


class SmppSessionState:
    """
    Represensts the states in which an SMPP session can be in.
    """

    # see section 2.2 of SMPP spec document v3.4
    # we are ignoring the other states since we are only concerning ourselves with an ESME in Transceiver mode.

    # An ESME has established a network connection to the SMSC but has not yet issued a Bind request.
    OPEN: str = "OPEN"
    # A connected ESME has requested to bind as an ESME Transceiver (by issuing a bind_transceiver PDU)
    # and has received a response from the SMSC authorising its Bind request.
    BOUND_TRX: str = "BOUND_TRX"
    # An ESME has unbound from the SMSC and has closed the network connection. The SMSC may also unbind from the ESME.
    CLOSED: str = "CLOSED"


class SmppCommand:
    """
    Represensts the various SMPP commands.
    """

    # see section 4 of SMPP spec document v3.4
    BIND_TRANSCEIVER: str = "bind_transceiver"
    BIND_TRANSCEIVER_RESP: str = "bind_transceiver_resp"
    BIND_TRANSMITTER: str = "bind_transmitter"
    BIND_RECEIVER: str = "bind_receiver"
    UNBIND: str = "unbind"
    UNBIND_RESP: str = "unbind_resp"
    SUBMIT_SM: str = "submit_sm"
    SUBMIT_SM_RESP: str = "submit_sm_resp"
    DELIVER_SM: str = "deliver_sm"
    DELIVER_SM_RESP: str = "deliver_sm_resp"
    ENQUIRE_LINK: str = "enquire_link"
    ENQUIRE_LINK_RESP: str = "enquire_link_resp"
    GENERIC_NACK: str = "generic_nack"

    # naz currently does not handle the following smpp commands.
    # open a github issue if you use naz and require support of a command in this list
    BIND_RECEIVER_RESP: str = "bind_receiver_resp"
    BIND_TRANSMITTER_RESP: str = "bind_transmitter_resp"
    QUERY_SM: str = "query_sm"
    QUERY_SM_RESP: str = "query_sm_resp"
    REPLACE_SM: str = "replace_sm"
    REPLACE_SM_RESP: str = "replace_sm_resp"
    CANCEL_SM: str = "cancel_sm"
    CANCEL_SM_RESP: str = "cancel_sm_resp"
    SUBMIT_MULTI: str = "submit_multi"
    SUBMIT_MULTI_RESP: str = "submit_multi_resp"
    OUTBIND: str = "outbind"
    ALERT_NOTIFICATION: str = "alert_notification"
    DATA_SM: str = "data_sm"
    DATA_SM_RESP: str = "data_sm_resp"
    RESERVED_A: str = "reserved_a"
    RESERVED_B: str = "reserved_b"
    RESERVED_C: str = "reserved_c"
    RESERVED_D: str = "reserved_d"
    RESERVED_E: str = "reserved_e"
    RESERVED_F: str = "reserved_f"
    RESERVED_G: str = "reserved_g"
    RESERVED_LIST_A: str = "reserved_list_a"
    RESERVED_LIST_B: str = "reserved_list_b"
    RESERVED_LIST_C: str = "reserved_list_c"
    RESERVED_LIST_D: str = "reserved_list_d"
    RESERVED_LIST_E: str = "reserved_list_e"
    RESERVED_LIST_F: str = "reserved_list_f"
    RESERVED_LIST_G: str = "reserved_list_g"
    RESERVED_LIST_H: str = "reserved_list_h"
    RESERVED_LIST_I: str = "reserved_list_i"
    RESERVED_FOR_SMPP_EXTENSION_A: str = "reserved_for_smpp_extension_a"
    RESERVED_FOR_SMPP_EXTENSION_B: str = "reserved_for_smpp_extension_b"
    RESERVED_FOR_SMSC_VENDOR_A: str = "reserved_for_smsc_vendor_a"
    RESERVED_FOR_SMSC_VENDOR_B: str = "reserved_for_smsc_vendor_b"


class CommandStatus(typing.NamedTuple):
    """
    An SMPP command status
    """

    code: str
    value: typing.Union[int, typing.List[int]]
    description: str


class SmppCommandStatus:
    """
    Represensts the various SMPP commands statuses.
    """

    # see section 5.1.3 of smpp ver 3.4 spec document

    ESME_ROK: CommandStatus = CommandStatus(
        code="ESME_ROK", value=0x00000000, description="Success"
    )
    ESME_RINVMSGLEN: CommandStatus = CommandStatus(
        code="ESME_RINVMSGLEN", value=0x00000001, description="Message Length is invalid"
    )
    ESME_RINVCMDLEN: CommandStatus = CommandStatus(
        code="ESME_RINVCMDLEN", value=0x00000002, description="Command Length is invalid"
    )
    ESME_RINVCMDID: CommandStatus = CommandStatus(
        code="ESME_RINVCMDID", value=0x00000003, description="Invalid Command ID"
    )
    ESME_RINVBNDSTS: CommandStatus = CommandStatus(
        code="ESME_RINVBNDSTS",
        value=0x00000004,
        description="Incorrect BIND Status for given command",
    )
    ESME_RALYBND: CommandStatus = CommandStatus(
        code="ESME_RALYBND", value=0x00000005, description="ESME Already in Bound State"
    )
    ESME_RINVPRTFLG: CommandStatus = CommandStatus(
        code="ESME_RINVPRTFLG", value=0x00000006, description="Invalid Priority Flag"
    )
    ESME_RINVREGDLVFLG: CommandStatus = CommandStatus(
        code="ESME_RINVREGDLVFLG", value=0x00000007, description="Invalid Registered Delivery Flag"
    )
    ESME_RSYSERR: CommandStatus = CommandStatus(
        code="ESME_RSYSERR", value=0x00000008, description="System Error"
    )
    ESME_RINVSRCADR: CommandStatus = CommandStatus(
        code="ESME_RINVSRCADR", value=0x0000000A, description="Invalid Source Address"
    )
    ESME_RINVDSTADR: CommandStatus = CommandStatus(
        code="ESME_RINVDSTADR", value=0x0000000B, description="Invalid Dest Addr"
    )
    ESME_RINVMSGID: CommandStatus = CommandStatus(
        code="ESME_RINVMSGID", value=0x0000000C, description="Message ID is invalid"
    )
    ESME_RBINDFAIL: CommandStatus = CommandStatus(
        code="ESME_RBINDFAIL", value=0x0000000D, description="Bind Failed"
    )
    ESME_RINVPASWD: CommandStatus = CommandStatus(
        code="ESME_RINVPASWD", value=0x0000000E, description="Invalid Password"
    )
    ESME_RINVSYSID: CommandStatus = CommandStatus(
        code="ESME_RINVSYSID", value=0x0000000F, description="Invalid System ID"
    )
    ESME_RCANCELFAIL: CommandStatus = CommandStatus(
        code="ESME_RCANCELFAIL", value=0x00000011, description="Cancel SM Failed"
    )
    ESME_RREPLACEFAIL: CommandStatus = CommandStatus(
        code="ESME_RREPLACEFAIL", value=0x00000013, description="Replace SM Failed"
    )
    ESME_RMSGQFUL: CommandStatus = CommandStatus(
        code="ESME_RMSGQFUL", value=0x00000014, description="Message Queue Full"
    )
    ESME_RINVSERTYP: CommandStatus = CommandStatus(
        code="ESME_RINVSERTYP", value=0x00000015, description="Invalid Service Type"
    )
    ESME_RINVNUMDESTS: CommandStatus = CommandStatus(
        code="ESME_RINVNUMDESTS", value=0x00000033, description="Invalid number of destinations"
    )
    ESME_RINVDLNAME: CommandStatus = CommandStatus(
        code="ESME_RINVNUMDESTS", value=0x00000034, description="Invalid Distribution List name"
    )
    ESME_RINVDESTFLAG: CommandStatus = CommandStatus(
        code="ESME_RINVDESTFLAG",
        value=0x00000040,
        description="Destination flag is invalid (submit_multi)",
    )
    ESME_RINVSUBREP: CommandStatus = CommandStatus(
        code="ESME_RINVSUBREP",
        value=0x00000042,
        description="Invalid (submit with replace) request(i.e. submit_sm with replace_if_present_flag set)",
    )
    ESME_RINVESMCLASS: CommandStatus = CommandStatus(
        code="ESME_RINVESMCLASS", value=0x00000043, description="Invalid esm_class field data"
    )
    ESME_RCNTSUBDL: CommandStatus = CommandStatus(
        code="ESME_RCNTSUBDL", value=0x00000044, description="Cannot Submit to Distribution List"
    )
    ESME_RSUBMITFAIL: CommandStatus = CommandStatus(
        code="ESME_RSUBMITFAIL", value=0x00000045, description="Submit_sm or submit_multi failed"
    )
    ESME_RINVSRCTON: CommandStatus = CommandStatus(
        code="ESME_RINVSRCTON", value=0x00000048, description="Invalid Source address TON"
    )
    ESME_RINVSRCNPI: CommandStatus = CommandStatus(
        code="ESME_RINVSRCNPI", value=0x00000049, description="Invalid Source address NPI"
    )
    ESME_RINVDSTTON: CommandStatus = CommandStatus(
        code="ESME_RINVDSTTON", value=0x00000050, description="Invalid Destination address TON"
    )
    ESME_RINVDSTNPI: CommandStatus = CommandStatus(
        code="ESME_RINVDSTNPI", value=0x00000051, description="Invalid Destination address NPI"
    )
    ESME_RINVSYSTYP: CommandStatus = CommandStatus(
        code="ESME_RINVSYSTYP", value=0x00000053, description="Invalid system_type field"
    )
    ESME_RINVREPFLAG: CommandStatus = CommandStatus(
        code="ESME_RINVREPFLAG", value=0x00000054, description="Invalid replace_if_present flag"
    )
    ESME_RINVNUMMSGS: CommandStatus = CommandStatus(
        code="ESME_RINVNUMMSGS", value=0x00000055, description="Invalid number of messages"
    )
    ESME_RTHROTTLED: CommandStatus = CommandStatus(
        code="ESME_RTHROTTLED",
        value=0x00000058,
        description="Throttling error (ESME has exceeded allowed message limits)",
    )
    ESME_RINVSCHED: CommandStatus = CommandStatus(
        code="ESME_RINVSCHED", value=0x00000061, description="Invalid Scheduled Delivery Time"
    )
    ESME_RINVEXPIRY: CommandStatus = CommandStatus(
        code="ESME_RINVEXPIRY",
        value=0x00000062,
        description="Invalid message validity period (Expiry time)",
    )
    ESME_RINVDFTMSGID: CommandStatus = CommandStatus(
        code="ESME_RINVDFTMSGID",
        value=0x00000063,
        description="Predefined Message Invalid or Not Found",
    )
    ESME_RX_T_APPN: CommandStatus = CommandStatus(
        code="ESME_RX_T_APPN",
        value=0x00000064,
        description="ESME Receiver Temporary App Error Code",
    )
    ESME_RX_P_APPN: CommandStatus = CommandStatus(
        code="ESME_RX_P_APPN",
        value=0x00000065,
        description="ESME Receiver Permanent App Error Code",
    )
    ESME_RX_R_APPN: CommandStatus = CommandStatus(
        code="ESME_RX_R_APPN",
        value=0x00000066,
        description="ESME Receiver Reject Message Error Code",
    )
    ESME_RQUERYFAIL: CommandStatus = CommandStatus(
        code="ESME_RQUERYFAIL", value=0x00000067, description="query_sm request failed"
    )
    ESME_RINVOPTPARSTREAM: CommandStatus = CommandStatus(
        code="ESME_RINVOPTPARSTREAM",
        value=0x000000C0,
        description="Error in the optional part of the PDU Body.",
    )
    ESME_ROPTPARNOTALLWD: CommandStatus = CommandStatus(
        code="ESME_ROPTPARNOTALLWD", value=0x000000C1, description="Optional Parameter not allowed"
    )
    ESME_RINVPARLEN: CommandStatus = CommandStatus(
        code="ESME_RINVPARLEN", value=0x000000C2, description="Invalid Parameter Length."
    )
    ESME_RMISSINGOPTPARAM: CommandStatus = CommandStatus(
        code="ESME_RMISSINGOPTPARAM",
        value=0x000000C3,
        description="Expected Optional Parameter missing",
    )
    ESME_RINVOPTPARAMVAL: CommandStatus = CommandStatus(
        code="ESME_RINVOPTPARAMVAL",
        value=0x000000C4,
        description="Invalid Optional Parameter Value",
    )

    ESME_RDELIVERYFAILURE: CommandStatus = CommandStatus(
        code="ESME_RDELIVERYFAILURE",
        value=0x000000FE,
        description="Delivery Failure (used for data_sm_resp)",
    )
    ESME_RUNKNOWNERR: CommandStatus = CommandStatus(
        code="ESME_RUNKNOWNERR", value=0x000000FF, description="Unknown Error"
    )
    RESERVED_A: CommandStatus = CommandStatus(
        code="Reserved", value=0x00000009, description="Reserved"
    )
    RESERVED_B: CommandStatus = CommandStatus(
        code="Reserved", value=0x00000010, description="Reserved"
    )
    RESERVED_C: CommandStatus = CommandStatus(
        code="Reserved", value=0x00000012, description="Reserved"
    )
    RESERVED_D: CommandStatus = CommandStatus(
        code="Reserved", value=0x00000041, description="Reserved"
    )
    RESERVED_E: CommandStatus = CommandStatus(
        code="Reserved", value=0x00000052, description="Reserved"
    )
    RESERVED_LIST_A: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000016, 0x00000032], description="Reserved"
    )
    RESERVED_LIST_B: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000035, 0x0000003F], description="Reserved"
    )
    RESERVED_LIST_C: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000046, 0x00000047], description="Reserved"
    )
    RESERVED_LIST_D: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000056, 0x00000057], description="Reserved"
    )
    RESERVED_LIST_E: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000059, 0x00000060], description="Reserved"
    )
    RESERVED_LIST_F: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000068, 0x000000BF], description="Reserved"
    )
    RESERVED_LIST_G: CommandStatus = CommandStatus(
        code="Reserved", value=[0x000000C5, 0x000000FD], description="Reserved"
    )
    RESERVED_LIST_H: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000100, 0x000003FF], description="Reserved for SMPP extension"
    )
    RESERVED_LIST_I: CommandStatus = CommandStatus(
        code="Reserved",
        value=[0x00000400, 0x000004FF],
        description="Reserved for SMSC vendor specific errors",
    )
    RESERVED_LIST_J: CommandStatus = CommandStatus(
        code="Reserved", value=[0x00000500, 0xFFFFFFFF], description="Reserved"
    )


class DataCoding(typing.NamedTuple):
    """
    An SMPP data encoding.
    """

    code: str
    value: int
    description: str


class SmppDataCoding:
    """
    Represensts the various SMPP data encodings.
    """

    # see section 5.2.19 of smpp ver 3.4 spec document.
    # also see:
    #   1. https://github.com/praekelt/vumi/blob/767eac623c81cc4b2e6ea9fbd6a3645f121ef0aa/vumi/transports/smpp/processors/default.py#L260
    #   2. https://docs.python.org/3/library/codecs.html
    #   3. https://docs.python.org/3/library/codecs.html#standard-encodings

    # The attributes of this class are equivalent to some of the names found in the python standard-encodings documentation
    # ie; https://docs.python.org/3/library/codecs.html#standard-encodings

    gsm0338: DataCoding = DataCoding(
        code="gsm0338", value=0b00000000, description="SMSC Default Alphabet"
    )
    ascii: DataCoding = DataCoding(
        code="ascii", value=0b00000001, description="IA5(CCITT T.50) / ASCII(ANSI X3.4)"
    )
    octet_unspecified_I: DataCoding = DataCoding(
        code="octet_unspecified_I",
        value=0b00000010,
        description="Octet unspecified(8 - bit binary)",
    )
    latin_1: DataCoding = DataCoding(
        code="latin_1", value=0b00000011, description="Latin 1 (ISO - 8859 - 1)"
    )
    octet_unspecified_II: DataCoding = DataCoding(
        code="octet_unspecified_II",
        value=0b00000100,
        description="Octet unspecified(8 - bit binary)",
    )
    # iso2022_jp, iso2022jp and iso-2022-jp are aliases
    # see: https://stackoverflow.com/a/43240579/2768067
    iso2022_jp: DataCoding = DataCoding(
        code="iso2022_jp", value=0b00000101, description="JIS(X 0208 - 1990)"
    )
    iso8859_5: DataCoding = DataCoding(
        code="iso8859_5", value=0b00000110, description="Cyrllic(ISO - 8859 - 5)"
    )
    iso8859_8: DataCoding = DataCoding(
        code="iso8859_8", value=0b00000111, description="Latin / Hebrew(ISO - 8859 - 8)"
    )
    # see: https://stackoverflow.com/a/14488478/2768067
    utf_16_be: DataCoding = DataCoding(
        code="utf_16_be", value=0b00001000, description="UCS2(ISO / IEC - 10646)"
    )
    ucs2: DataCoding = DataCoding(
        code="ucs2", value=0b00001000, description="UCS2(ISO / IEC - 10646)"
    )
    shift_jis: DataCoding = DataCoding(
        code="shift_jis", value=0b00001001, description="Pictogram Encoding"
    )
    iso2022jp: DataCoding = DataCoding(
        code="iso2022jp", value=0b00001010, description="ISO - 2022 - JP(Music Codes)"
    )
    # reservedI= DataCoding(code="reservedI", value=0b00001011, description= "reserved")
    # reservedII= DataCoding(code="reservedII", value=0b00001100, description= "reserved")
    euc_kr: DataCoding = DataCoding(code="euc_kr", value=0b00001110, description="KS C 5601")

    # not the same as iso2022_jp but ... ¯\_(ツ)_/¯
    # iso-2022-jp=DataCoding(code="iso-2022-jp", value=0b00001101, description="Extended Kanji JIS(X 0212 - 1990)")

    # 00001111 - 10111111 reserved
    # 0b1100xxxx GSM MWI control - see [GSM 03.38]
    # 0b1101xxxx GSM MWI control - see [GSM 03.38]
    # 0b1110xxxx reserved
    # 0b1111xxxx GSM message class control - see [GSM 03.38]


class SmppOptionalTag:
    """
    Represensts the various SMPP Optional Parameter Tags.
    """

    # see section 5.3.2 of smpp ver 3.4 spec document.
    # All optional parameters have the following general TLV (Tag, Length, Value) format.
    # Tag, Integer, 2octets
    # Length, Integer, 2octets
    # Value, type varies, size varies. eg receipted_message_id is of type c-octet string of size 1-65

    dest_addr_subunit = 0x0005
    dest_network_type = 0x0006
    dest_bearer_type = 0x0007
    dest_telematics_id = 0x0008
    source_addr_subunit = 0x000D
    source_network_type = 0x000E
    source_bearer_type = 0x000F
    source_telematics_id = 0x0010
    qos_time_to_live = 0x0017
    payload_type = 0x0019
    additional_status_info_text = 0x001D
    receipted_message_id = 0x001E
    ms_msg_wait_facilities = 0x0030
    privacy_indicator = 0x0201
    source_subaddress = 0x0202
    dest_subaddress = 0x0203
    user_message_reference = 0x0204
    user_response_code = 0x0205
    source_port = 0x020A
    destination_port = 0x020B
    sar_msg_ref_num = 0x020C
    language_indicator = 0x020D
    sar_total_segments = 0x020E
    sar_segment_seqnum = 0x020F
    SC_interface_version = 0x0210
    callback_num_pres_ind = 0x0302
    callback_num_atag = 0x0303
    number_of_messages = 0x0304
    callback_num = 0x0381
    dpf_result = 0x0420
    set_dpf = 0x0421
    ms_availability_status = 0x0422
    network_error_code = 0x0423
    message_payload = 0x0424
    delivery_failure_reason = 0x0425
    more_messages_to_send = 0x0426
    message_state = 0x0427
    ussd_service_op = 0x0501
    display_time = 0x1201
    sms_signal = 0x1203
    ms_validity = 0x1204
    alert_on_message_delivery = 0x130C
    its_reply_type = 0x1380
    its_session_info = 0x1383
